/*-----------------------------------------------------------------------------
FILE    : crsapply.c
CONTENT : apply hierarchical model of condensed random set annotations and 
          compute sample probabilities and/or database log-likelihood measures
          crsapply reads saved hierarchical condensed random set model (see 
          crsinduce for model induction) from file and assess the fit of and 
          instantiation database to that model. It computes summary statistics
          for the fit of the database as a whole and can optionally provide 
          log-likelihood scores for individual instantiations.  
          crsapply is the command line interface to query models implemented by
          the hdist package  
Author  : Frank Rgheimer
History : 2008-12-?? core elements created within psvmodel 
          ---------------------------------------------------------------------
          2012-02-19 branched from psvmodel/crsinduce as part of modular 
                     interface; basic file structure
          2012-02-29 updated command line interface and help screen
                     wrote main program and adapted evaluation form psvmodel
                     checked for memory leaks -- none found
          2012-04-23 added command line support for frame specification file
          2012-04-24 successfully tested argument processing for -f
                     added treatment of frame specifications 
          2012-05-17 adapted code to suppress warning triggered by changes in
                     hdist function prototypes for stricter type checking 
           
TODO:     check compatibility with non-default interpretations 
          write unit tests
-------------------------------------------------------------------------------
 Copyright 2012 Frank Rgheimer
 This program is distributed under the terms of the GNU Lesser General Public
 License.
-----------------------------------------------------------------------------*/

#include "crsutil.h"              /* pull in all headers and definitions  */

/* --- program-specific definitions --- */
#define MAXFRAMESIZE 2048          /* maximum number of nodes in frame spec.*/ 
/* increase value to allow for larger frames */

#define DESCRIPTION "apply/evaluate hierarchical condensed random set model"
#define VERSION "\
version 0.1.0  (2012-02-22)"  
#define COPYRIGHTS        "          Copyright 2012      Frank Rgheimer\n\
attmap, tabscan and parser modules   Copyright 1996-2010 Christian Borgelt\n\
hdist module                         Copyright 2007-201O, 2012\n\
                                                         Frank Rgheimer\n\
This program may be distributed under the terms of the GNU Lesser General \n\
Public License."

/*---globals---*/
/* static char   *progname; */
int	       options=0;         /* for option flags, e.g. set-valued eval */
static TABSCAN *scanner= NULL;    /* scanner for table file                 */
static SCAN    *scan   = NULL;    /* general scanner                        */  
static HDIST   *hd     = NULL;    /* for representation of distributions    */
static FILE    *datfile= NULL;    /* ptr to a data file which comprises a  
                   single header line header with <space>-separated attribute
                   names and one line per record with <space>-separated att-
                   ribute values. Used as training data for model induction */
char withheader= FALSE;      /* datfile has header        (default: no)     */
double lcorr=0.0;            /* laplace correction default value            */
static FILE *mdlfile=NULL;   /* hdist model file (written via hdist_save()  */
static FILE *outfile=NULL;   /* write assessed table model to this file     */
static FILE *frmfile=NULL;   /* file with optional frame specification      */

static const MINFO  modes[]={         /* registered option arguments for -m */
                {HDBM_NXHTV,"disj"},  /* see help for documentation         */
                {HDBM_NXHTV,"incl"},
                {HDBM_NXHTV,"other"},
                {HDBM_IMPR, "conj"},
                {HDBM_IMPR, "impr"},
                {HDBM_XHTV, "split"},
                {HDBM_XHTV, "uncert"},
                {HDBM_XHTV, "expand"},
                {-1,NULL}             /* sentinel value */
             };
  

/*-----------------------------------------------------------------------------
    obligatory implementations
-----------------------------------------------------------------------------*/
/** clean up any global objects and set pointers to NULL to mark them as
  deleted  */ 
void cleanup(){
  if (scanner) {ts_delete(scanner); scanner=NULL;}
  if (scan) {sc_delete(scan); scan=NULL;}
  if (hd) {hdist_delete(hd); hd=NULL;}
  if (mdlfile)  fclose(mdlfile); mdlfile=NULL;
  if (datfile) fclose(datfile); datfile=NULL;
  if (outfile)  fclose(outfile); outfile=NULL;
  if (frmfile) fclose(frmfile); frmfile=NULL;
}

/*===========================================================================*/

/** generic greeting message and copyright statement */
static void greetings(){
  fprintf(stderr,"%s - %s\n",progname, DESCRIPTION);
  fprintf(stderr,"%s%s\n"   , VERSION, COPYRIGHTS);
}


/* --help screen--
   text layout 75 characters wide for display in standard terminals */
static void help (void){
  fprintf(stderr,"\n"\
  "\nUSAGE: %s [options] mdlfile [datfile [outfile]]\n", progname);
  fprintf(stderr,
  "\nDESCRIPTION:"\
  "\n  This program reads a hierarchical distribution model and a database of"\
  "\n  (set-)instantiations and assesses the fit of the data to the model."\
  "\n  If outfile is specified log-likelihoods for individual instantiations"\
  "\n  are written to that file. The program writes a table with fit statis-"\
  "\n  tics of the database as a whole to stdout"); 
    fprintf(stderr,"\nARGUMENTS:" 
  "\nmdlfile hdist model file (e.g. from crsinduce)"\
  "\ndatfile database with (set instantiations) to be assessed"
  "\n        Format description follows:"); 
  fprintf(stderr,
  "\n        Each table has an optional comment part (line starting with '#'"\
  "\n        followed by a header line with field names. Data records take"\
  "\n        one each, with fields separated by tabulators.");
  fprintf(stderr,
  "\n        Any field may either contain an identifier or a set of identifi-"\
  "\n        ers that are separated by spaces and enclosed in curly brackets"\
  "\n        (e.g. {A B C}). "\
  "\n        By default, the last field is assumed to contain the attribute"\
  "\n        to be processed. If datfile is not specified it defaults to "\
  "\n        stdin.");
  fprintf(stderr,
  "\noutfile If outfile is specified a table associating any valid (set-)"\
  "\n        instantiation from datfile  along with their respective log-"\
  "\n        likelihood under the model specified in mdlfile will be written"\
  "\n        to that file. The interpretation of non-leaf symbols and set-"\
  "\n        tings for Laplace corrections follow specifications in mdlfile");
  fprintf(stderr,"\nOPTIONS:"
/* "\n  -i    treat the first non-comment record of the input as a header"\
  "\n        (by default all records are considered as data)"\ */
/*  "\n  -l#.# set value for laplace correction to #.#      (default=0.0)"\
  "\n  -m<mode> set interpretation mode for non-leaf nodes; options are:"\
  "\n        disj, incl or other: non-leaf node indicate outcome disjoint"\
  "\n          from the possible expansions of the node this is equivalent");
  fprintf(stderr,
  "\n          to having an implicit \"other\" outcome in each branch that"\
  "\n          can absorb probability mass and contributes to the coverage"\
  "\n          rate of the branch;                 (default configuration)"\
  "\n          distributions saved in this mode can also serve as interme-"\
  "\n          diates if delayed expnasion to any of the other interpreta-"\
  "\n          tions is desired");
  fprintf(stderr,
  "\n        conj or impr: non-leaf nodes serve as placeholders for *all*"\
  "\n          possible expansions of that node -- used, e.g., as imprecise"\
  "\n          specification."\
  "\n        expand, split or uncert: force expansion of outcome tree after"\
  "\n          training. Probabiltiy mass and coverage contributions are");
  fprintf(stderr,
  "\n          split according to the distribution of expansions of the node"\
  "\n          observed in the remaining data. This mode is consistent with a"\
  "\n          model that permits only leave nodes and sets of leave nodes as"\
  "\n          outcomes"); */
  "\n  -f <framefile>   read frame specification from framefile -- nodes "\
  "\n        listed in framefile will be considered as leaves for query in-"\
  "\n        terpretation."\
  "\n  -s    silent mode (suppress greeting and copyright message at startup)"\
  "\n");
  return;
}

/*---------------------------------------------------------------------------*/

/* modified version of test_hd function of psvmodel */
/** evaluation of a test database against a hdist model 
    @arg hd hierarchical distribution model
    @arg tabscan table file scanner from util package 
    @arg tstdata sample database to evaluate each sample consists of a list of 
          node identifiers that represent a set instantiation w.r.t. any of
          the frames generated by hd's underlying the hierarchy. Lists may 
          optionally be enclosed with curly braces. An empty pair of curly 
          braces {} or { } represents the empty set.  Note that empty input 
          lines are ignored by the parser, i.e. braces are required to refer
          to empty instantiations.
    @arg file outfile if different from NULL evaluations for each individual
          sample read from testdata will be written here  */
int eval_db(HDIST* hd, TABSCAN* scanner, FILE* tstdata, FILE* outfile){
  int      i,l,d,maxel,elcnt,reccnt;
  char     buffer[2048];       /* reading buffer for parser */
  int      bufsz=2048;
  char*    valnames[100];      /* for storing string pointers */
  char**   p;                  /* for traversing token stings */ 
  char*    token;
  char*    errtoken;           /* token for error message                   */
  double   logp_test;          /* to aggregate log-likelihood of test data  */
  double   logp_smpl;          /* log likelihood of current sample          */
  double   logp_msmpl;         /* mean log likelihood of test data sample   */
  double   qerrfrq;            /* squared error sum (singleton frequencies) */
  double   qerrcov;            /* squared error sum (coverage rates)        */
  double*  frq_cov_o  =NULL;   /* observed element coverage rates in sample */
  double*  frq_sglt_o =NULL;   /* observed frequency of singleton outcomes  */
  ATT*     att;
  HD_NODE* pn;
 
  assert(hd&&scanner&&tstdata);
  att=hd->att; assert(att);     
  maxel=(sizeof(valnames)/sizeof(char*))-1;       /* =99 */
  frq_cov_o=(double*)calloc(att->valcnt,sizeof(double));
  frq_sglt_o=(double*)calloc(att->valcnt,sizeof(double));
  if (!(frq_cov_o&&frq_sglt_o)) error(E_NOMEM);
  logp_test=0.0;               /* test with no sample always ok */
  reccnt=0;
  while ((d=readnext_sv(buffer,bufsz,valnames,maxel,scanner,tstdata))){
    if (d<0) break;            /* exit loop with error code as return value */
    errtoken=NULL; 
    /* we still need to parse, but frequencies are retrieved from the model */
    /* update log likelihood and observed freq/coverage                     */   
    logp_smpl=hdist_setlogl(hd,(const char**) valnames, maxel, 
                               (const char**) &errtoken,0);
    if (errtoken){
       fprintf(stderr,
         "Error: Invalid set specification -%s%s-."
         "\nPlease check the model and query files!\n",
         *errtoken?"term ":"",errtoken);
       exit(1);    
    }
    logp_test+=logp_smpl;
    elcnt=0; reccnt++;
    if (outfile) fprintf(outfile, "{"); 
    for (p=valnames;*p;elcnt++) {
      token=*p; p++;
      l=att_valid(att,token);
      if (l<0) {                /* skip warning on empty token */
        fprintf(stderr,"warning: ignoring unknown attribute value: -%s-\n",token);
      }
      else {
        if (outfile) { 
          fprintf(outfile," %s%s",token,*p?"":" }");
          if (!*p) fprintf(outfile,"\t%f\n",logp_smpl);
        }
        frq_cov_o[l]++;          /* update coverage and singleton counters */
        if (elcnt==0 && !(*p)) frq_sglt_o[l]++;    /* cond. for singleton  */
      }
    }
    if (elcnt==0) {              /* treat empty set special case */
      if (outfile) fprintf(outfile," }\t%f\n",logp_smpl);
    }
  }
  logp_msmpl=(reccnt>0)?(logp_test/reccnt):(1.0);

  /* compute relative value freq and coverage rate in test data (no lcorr) */
  qerrfrq=qerrcov=0.0;
  for (i=0;(reccnt>0)&&(i<att->valcnt);i++){
     frq_sglt_o[i]=(frq_sglt_o[i]/reccnt);
     frq_cov_o[i]=(frq_cov_o[i]/reccnt);
     pn=hd->node[i]; assert(pn);
#if false
#ifndef NDEBUG
     printf("%s : (freq,obs)=(%f,%f), (cov,obs)=(%f,%f)\n",
        (att_valname(hd->att,i))?att_valname(hd->att,i):"n.a.",
        hdist_nodefreq(hd,pn),frq_sglt_o[i],
        hdist_nodecov(hd,pn),frq_cov_o[i]);
#endif
#endif
     qerrfrq+=pow( (hdist_nodefreq(hd,pn)-frq_sglt_o[i]) ,2.0 );
     qerrcov+=pow( (hdist_nodecov(hd,pn)-frq_cov_o[i]) ,2.0 );
  }
  printf("\n-------------------------------------------------------------\n");
  printf("model\t logL(D_test)\tlogL(d_avg)\tsum err_frq^2\
\tsum err_cov^2\n");
  printf("-------------------------------------------------------------\n");
  printf("hdist\t%f\t%f\t%f\t%f\n",logp_test,logp_msmpl,qerrfrq,qerrcov);
  /*
  printf("hdist: \nlog-likelihood of test data &\t%f\n",logp_test);
  printf("err_{cov}^2 &\t%f\n",qerrcov);
  printf("err_{frq}^2 &\t%f\n",qerrfrq);
  */
  
  if (frq_cov_o)  free(frq_cov_o);
  if (frq_sglt_o) free(frq_sglt_o);
  return(errtoken?(-1):0); /* error status */
}

/*--------------------------------------------------------------------------*/

/** read a frame specification from file and store it in an array of node 
    identifiers
    @arg hd ptr to hierarchical distribution referenced
    @arg scanner table file scanner 
    @arg framefile read frame specifications from this file
    @return status code (0 for success) 
    @note currently using fixed buffer size -- replace by dynamic data structs
          if even larger frames are desired */
int readframe(HDIST* hd, TABSCAN* scanner, FILE* framefile){
   /* use readnext_sv to parse frame specification, then set flags in hd    */
  int      i, d, maxnodes;                  /* id, status code, node limit  */
  char     buffer[4096];                    /* reading buffer for parser    */
  int      bufsz=4096;
  char* nodenames[MAXFRAMESIZE];            /* for storing string pointers  */
  char**   p;                               /* for traversing token stings  */
  assert(hd && scanner && framefile);
  maxnodes=(sizeof(nodenames)/sizeof(char*))-1;/* max. # nodes in frame     */
  d=readnext_sv(buffer, bufsz, nodenames, maxnodes, scanner, framefile); 
  if (d<0) error(E_READFRAME);
  hdist_clearframe(hd);                  /* clear previous frame spec    */
  /* traverse string buffers and set node noexpand flags in hd              */
  for (p=nodenames;*p;p++){  
    i=hdist_nodeid(hd,*p);                  /* get numerical node id        */
    if (i<0) return (-1);                   /* unrecognized node id         */
    d=hdist_add2frame(hd,i);                /* prevent node expansion       */
    if (d) return (-1);                     /* unrecognized node identifier */
  }
  return 0;
}

/*--------------------------------------------------------------------------*/
/*---main function---*/

int  main(int argc, char* argv[]){
  int i, k=0;                      /* loop variable and argument counter    */
  /*int hdist_mode=HDBM_DISJ; */   /* default interpretation for hdist      */
  int stdincnt=0;                  /* count stdin assignments               */
  int silent=0;                    /* flag to suppress greeting in scripts  */
  char       *s;                   /* for scanning options                  */
  /*char       *lcoastr = NULL; */ /* the Laplace correction option argument*/
  /*char       *modestr = NULL; */ /* for setting mode on command line      */
  char       **optarg = NULL;      /* for redirecting  option arguments     */
  char       *fn_mdl  = NULL;      /* name of model file                    */
  char       *fn_dat  = "";        /* name of database file                 */
  char       *fn_out  = NULL;      /* name of output file                   */
  char       *fn_frm  = NULL;      /* name of frame specification file      */

  
  progname=argv[0];                /* get program name for messages */ 
  /* --- evaluate arguments --- */ 
  for (i=1; i< argc; i++) {        /* traverse the argument vector        */
    s=argv[i];                     /* get next argument */
    if (optarg) {                  /* if at start of an option argument   */
      *optarg=s;                   /* set resp. var to reading position   */
      optarg=NULL;                 /* doubles as an indicator tnx to C.B  */
      continue;                    /* read next command line argument     */
    }
    /* otherwise */
    if ((*s == '-') && *++s) {     /* for options ignore leading '-' */
      while (*s) {                 /* may be list of options */
        switch (*s++) {            /* evaluate options */
        case '?': help();                  return 0; 
        /* disable unused options
        case 'i': withheader=FALSE;        break; 
        case 'l': optarg=&lcoastr;         break;    
        case 'm': optarg=&modestr;         break;
           end of unused options */ 
        case 'f': optarg=&fn_frm;          break;
        case 's': silent=1;                break;
        
	default : error(E_OPTION,s-2);     return 0; /* unknown option error */
	}
        /* option arguments may follow directly or in the next cmdarg */
        if (!*s) break;            /* end of cmdarg - try next one  */
        if (optarg) {*optarg =s; optarg=NULL; break;} 
        /* identify rest of current cmdarg with the option argument */
      }
    }
    /* note: single dashes '-' have been consumed by the option test */
    else {                         /* process non-option arguments */
      switch (k++) {               /* match arguments by position */
                                   /* use 1st argument for domain file name */
        case  0: fn_mdl=s; if (!*s) stdincnt++; break;  /* "-" means stdin */
                                   /* use 2st argument for training file name*/ 
        case  1: fn_dat=s; if (!*s) stdincnt++; break; 
        case  2: fn_out=s; break;  /* and 3nd argument for output file name */
      default: error(E_ARGCNT); break; /* too many arguments */
      }
    }
  }
  if (optarg) error(E_OPTARG);     /* missing option argument */
  /* check if to read frame specification argument from stdin */
  if (fn_frm && (*fn_frm == '-')){ /* check if  file name starts with "-" */
     if (*(fn_frm+1)) error(E_OPTARGDSC_2,"-f",fn_frm); /* invalid name */
     stdincnt++; fn_frm++;         /* read over "-" */
  }
  /* if no arguments given at all -> show help 
                        otherwise -> show greetings */
  if (k >= 1) {                    /* if there are arguments given*/   
    if (!silent) greetings();      /* print greetings... */
  }                                /* for correct */
  else{                            /* ...otherwise...*/ 
    help();                        /* ...show help message */
    exit(2);                       /* and abort program */
  }
  /* post-process and check arguments ... */ 
  if (stdincnt>1) error(E_STDINCNT);
  if (k<1) error(E_ARGCNT);        /* are all required arguments present? */ 
  assert(fn_mdl);
  assert(fn_dat);
  /* extra code to indicate frame restrictions */
  if (fn_frm) fprintf(stderr,"reading frame restrictions from %s\n",fn_frm);           
  /* load model file into hd variable  */
  scan=sc_create(fn_mdl);
  if (!scan) {
    if (!fn_mdl) error (E_UNKNOWN);  /* should have been captured by assert */
    else error(E_OPENF,fn_mdl);
  }
  pa_init(scan);
  hd=hdist_load(scan);
  /* hdist_print(hd); */ 
  if (scan) sc_delete(scan); scan=NULL;   /* free scan */
  /* set up parser for data file */
  scanner=ts_create();
  if(!scanner) error(E_NOMEM);/* scanner initialization & config.*/
  ts_chars(scanner,TS_FLDSEP,",\t");/* set field separators, ...        */
  ts_chars(scanner,TS_RECSEP,"\n"); /* ...and record separator          */
  ts_chars(scanner,TS_COMMENT,"#"); /* set indicator for comment lines  */
  /* blank symbols serve to distinguish elements in sets */
  /* if a frame file is specified: read it and set the frame in hd */ 
  if (fn_frm) {
    if (!*fn_frm) frmfile=stdin;
    else if (!(frmfile=fopen(fn_frm,"r"))) error (E_OPENF,fn_frm);
    if (readframe(hd,scanner,frmfile)!=0) error (E_READFRAME); 
    ts_reset(scanner); /* reset the table scanner before reading next file */ 
    fprintf(stderr,"Current frame is: "); hdist_showframe(hd);  
  }
  /* open data and output files empty strings stand for stdin and stdout */
  if (!*fn_dat) datfile=stdin; 
  else if (!(datfile=fopen(fn_dat,"r"))) error (E_OPENF,fn_dat);
  if (fn_out) {
    if (!*fn_out) outfile=stdout;
    else if (!(outfile=fopen(fn_out,"w"))) error (E_OPENF,fn_out); 
  }
  fprintf(stderr,"\n...assessing data from file %s\n",fn_dat);
  if (eval_db(hd, scanner,datfile, outfile)!=0) error (E_PARSETBL,fn_dat);
  cleanup();                          /* does not work according to valgrind */
  return 0;
}
