/*-----------------------------------------------------------------------------
FILE    : crsinduce.c
CONTENT : induce hierarchical distribution for condensed random sets from data 
          crsinduce acts as command line interface to the model induction
          functionality in the hdist package  
Author  : Frank Rgheimer
History : 2012-02-14 branched from psvmodel to provide modular interface
          2012-02-18 moved code shared code to crsutil.h
          2012-02-19 enabled stdin as input file, made output file argument 
                     optional with default set to stdout
                     added -m option to select HDMF distribution types
                     Valgrind reports evidence of failure to free hd structure
                     during cleanup() procedure -- added protected free in 
                     main() as workaround.  -> v0.1.0
          2012-02-22 fixed original cause of memory leak (extra definition of
                     hd in local main() scope made structure inaccessible to 
                     cleanup() operation by masking global declaration of same 
                     name); fixed typos in help -> v1.0.0 
          2012-03-01 expanded help screen to clarify disj/incl interpretation
                     setting 
TODO:    add automatic test procedure with FRDEBUG (use test4.*)
         
-------------------------------------------------------------------------------
 Copyright 2012 Frank Rgheimer
 This program is distributed under the terms of the GNU Lesser General Public
 License.
-----------------------------------------------------------------------------*/

#include "crsutil.h"                  /* pull in all headers and definitions */

/* --- program-specific definitions --- */

#define DESCRIPTION "induce hierarchical condensed random set model"
#define VERSION "\
version 1.0.0  (2012-02-22)"  
#define COPYRIGHTS        "          Copyright 2012      Frank Rgheimer\n\
attmap, tabscan and parser modules   Copyright 1996-2010 Christian Borgelt\n\
hdist module                         Copyright 2007-201O, 2012\n\
                                                         Frank Rgheimer\n\
This program is distributed under the terms of the Gnu Lesser General Public\n\
License."

/*---globals---*/
/* static char   *progname; */
int	       options=0;         /* for option flags, e.g. set-valued eval */
static TABSCAN *scanner= NULL;    /* scanner for table file                 */
static SCAN    *scan   = NULL;    /* general scanner                        */  
static HDIST   *hd     = NULL;    /* for representation of distributions    */
static FILE    *datafile=NULL;    /* ptr to a data file which comprises a  
                   single header line header with <space>-separated attribute
                   names and one line per record with <space>-separated att-
                   ribute values. Used as training data for model induction */
char withheader= FALSE;      /* datafile has header        (default: no)    */
double lcorr=0.0;            /* laplace correction default value            */
static FILE *treefile=NULL;  /* used to read tree structure - format is subset
                                of graphviz dot language see example.htr only
                                required for models with structured domain  */
static FILE *outfile=NULL;   /* write induced model to this file            */

static const MINFO  modes[]={         /* registered option arguments for -m */
                {HDBM_NXHTV,"disj"},  /* see help for documentation */
                {HDBM_NXHTV,"incl"},
                {HDBM_NXHTV,"other"},
                {HDBM_IMPR, "conj"},
                {HDBM_IMPR, "impr"},
                {HDBM_XHTV, "split"},
                {HDBM_XHTV, "uncert"},
                {HDBM_XHTV, "expand"},
                {-1,NULL}             /* sentinel value */
             };
  

/*-----------------------------------------------------------------------------
    obligatory implementations
-----------------------------------------------------------------------------*/
/** clean up any global objects and set pointers to NULL to mark them as
  deleted  */ 
void cleanup(){
  if (scanner) {ts_delete(scanner); scanner=NULL;}
  if (scan) {sc_delete(scan); scan=NULL;}
  if (hd) {hdist_delete(hd); hd=NULL;}
  if (treefile) fclose(treefile); treefile=NULL;
  if (datafile) fclose(datafile); datafile=NULL;
  if (outfile)  fclose(outfile); outfile=NULL;
}

/*===========================================================================*/


static void greetings(){
  fprintf(stderr,"%s - %s\n",progname, DESCRIPTION);
  fprintf(stderr,"%s%s\n"   , VERSION, COPYRIGHTS);
}


/* --help screen--
   text layout 75 characters wide for display in standard terminals */
static void help (void){
  fprintf(stderr,"\n"\
  "\nUSAGE: %s [options] dotfile trnfile [outfile]\n", progname);
    fprintf(stderr,"\nARGUMENTS:" 
  "\ndotfile provides a hierarchy structure as a directed tree in simplified"\
  "\n        .dot format."\
  "\ntrnfile contains tables with training data. Format description follows:"); 
  fprintf(stderr,
  "\n        Each table has an optional comment part (line starting with '#'"\
  "\n        followed by a header line with field names. Data records take"\
  "\n        one each, with fields separated by tabulators.");
  fprintf(stderr,
  "\n        Any field may either contain an identifier or a set of identifi-"\
  "\n        ers that are separated by spaces and enclosed in curly brackets"\
  "\n        (e.g. {A B C}). "\
  "\n        By default, the last field is assumed to contain the attribute"\
  "\n        to be processed");
  fprintf(stderr,
  "\noutfile specifies the name of an output file for saving the induced mo-"\
  "\n        del. Output consists of several variable-value pairs for model"\
  "\n        parameters and an enriched tree structure annotated with the"\
  "\n        distribution information.\n");
  fprintf(stderr,"\nOPTIONS:"\
/* "\n  -i    treat the first non-comment record of the input as a header"\
  "\n        (by default all records are considered as data)"*/
  "\n  -l#.# set value for Laplace correction to #.#      (default=0.0)"\
  "\n  -m<mode> set interpretation mode for non-leaf nodes; options are:"\
  "\n        disj, incl or other: non-leaf node indicate non-specific"\
  "\n          outcome separate from possible expansions of the node.");
  fprintf(stderr,
  "\n          Note: This is similar to using explite \"other\" outcome in"\
  "\n          each branch. Unexpanded intermediate nodes can absorb probabi-"\
  "\n          lity mass and contribute towards the coverage rates of their"\
  "\n          respective sub-trees. However, that unlike with separate "\
  "\n          nodes, the generating grammar for instantiations over hierar-");
  fprintf(stderr,
  "\n          chical  domains prevent a node from occurring in the same set"
  "\n          outcome as any node from its expansion sub-tree. If such in-"\
  "\n          valid value combinations are encountered in instantiation "\
  "\n          specifications the (implicit) ancestor node will be silently"\
  "\n          dropped. Distributions saved in this mode can also serve as");
  fprintf(stderr,
  "\n          intermediates if delayed expansion to any of the other inter-"\
  "\n          pretations is desired. It is used e.g., for the Gene Ontology"\
  "\n          standard, which requires genes to be annotated with non-leaf"\
  "\n          terms when this constitutes the most specific characterization"\
  "\n          supported by the available evidence.  (default configuration)");
  fprintf(stderr,
  "\n        conj or impr: non-leaf nodes are expanded to the set of all leaf"\
  "\n          nodes in the subtree rooted at the respective node -- used,"\
  "\n          e.g., as an imprecise specification."\
  "\n        expand, split or uncert: force expansion of outcome tree after"\
  "\n          training. Probability mass and coverage contributions are");
  fprintf(stderr,
  "\n          split according to the distribution of expansions of the node"\
  "\n          observed in the remaining data. This mode is consistent with a"\
  "\n          model that permits only leave nodes and sets of leave nodes as"\
  "\n          outcomes"\
  "\n  -s    silent mode (suppress startup greeting and copyright message)"\
  "\n");
  return;
}

/*---------------------------------------------------------------------------*/


/*---main function---*/

int  main(int argc, char* argv[]){
  int i, k=0;                      /* loop variable and argument counter */
  int hdist_mode=HDBM_DISJ;        /* default interpretation for hdist   */
  int stdincnt=0;                  /* count stdin assignments */ 
  int silent=0;                    /* flag to supress greeting in scripts */
  char       *s;                   /* for scanning options */
  char       *lcoastr = NULL;      /* the laplace correction option argument*/
  char       *modestr = NULL;      /* for setting mode on command line */ 
  char       **optarg = NULL;      /* for redirecting  option arguments     */
  char       *fn_trn  = NULL;      /* name of table file used for training  */
  char       *fn_dom  = NULL;      /* name of base attribute domain file    */
  char       *fn_out  = "";        /* name of output file (default: stdout) */ 
  /*ATT*       att      = NULL;    */  /* base attribute */
  
  progname=argv[0];                /* get program name for messages */ 
  /* --- evaluate arguments --- */ 
  for (i=1; i< argc; i++) {        /* traverse the argument vector   */
    s=argv[i];                     /* get next argument */
    if (optarg) {                  /* if at start of an option argument  */
      *optarg=s;                   /* set resp. var to reading position  */
      optarg=NULL;                 /* doubles as an indicator tnx to C.B */
      continue;                    /* read next command line argument    */
    }
    /* otherwise */
    if ((*s == '-') && *++s) {     /* for options ignore leading '-' */
      while (*s) {                 /* may be list of options */
        switch (*s++) {            /* evaluate options */
        case '?': help();                  return 0; 
        case 'l': optarg=&lcoastr;         break;    
        case 'm': optarg=&modestr;         break;
        case 's': silent=1;                break;
        /* case 'i': withheader=FALSE;      break;   deactivated */ 
	default : error(E_OPTION,s-2);     return 0; /* unknown option error */
	}
        /* option arguments may follow directly or in the next cmdarg */
        if (!*s) break;            /* end of cmdarg - try next one  */
        if (optarg) {*optarg =s; optarg=NULL; break;} 
        /* identify rest of current cmdarg with the option argument */
      }
    }
    /* note: single dashes '-' have been consumed by the option test */
    else {                         /* process non-option arguments */
      switch (k++) {               /* match arguments by position */
                                   /* use 1st argument for domain file name */
        case  0: fn_dom=s; if (!*s) stdincnt++; break;  /* "-" means stdin */
                                   /* use 2st argument for training file name*/ 
        case  1: fn_trn=s; if (!*s) stdincnt++; break; 
        case  2: fn_out=s; break;  /* and 3nd argument for output file name */
      default: error(E_ARGCNT); break; /* too many arguments */
      }
    }
  }
  if (optarg) error(E_OPTARG);     /* missing option argument */
  /* match option argument strings to internal flags */
  if (lcoastr!=NULL) {              /* parse value for laplace correction */
    lcorr=strtod(lcoastr,NULL);
    if (errno||(lcorr<0.0)) error(E_LCORR);
  }
  if (modestr!=NULL) {
    if ((i=code(modes,modestr))>=0) hdist_mode=i;
    /* fprintf(stderr,"%i , %s\n",i,modestr); */
  }
  /* if no arguments given at all -> show help 
                        otherwise -> show greetings */
  if (argc > 1) {                  /* if there are arguments given*/   
    if (!silent) greetings();      /* print greetings... */
  }                                /* for correct */
  else{                            /* ...otherwise...*/ 
    help();                        /* ...show help message */
    exit(2);                       /* and abort program */
  }
  /* post-process and check arguments ... */ 
  if (stdincnt>1) error(E_STDINCNT);
  if (k<2) error(E_ARGCNT);        /* are all required arguments present? */ 
  assert(fn_dom);
  assert(fn_trn);
  assert(fn_out);
 
  /* read domain file and create attribute */
  scan=sc_create(fn_dom);  if (!scan) {
    if (!fn_dom||!*fn_dom) error (E_NOMEM);
    else error(E_OPENF,fn_dom);
  }
  /* add option for alternative  mode flags */
  pa_init(scan);                      /* use parser ext. for reading hdist */
  hd=hdist_parse(scan,hdist_mode);    /* read the domain hierarchy */
  if (!hd) error(E_PARSETREE);
  if (scan) sc_delete(scan); scan = NULL;   /* safely delete the scanner   */
  /* process input table */
  scanner=ts_create();
  if(!scanner) error(E_NOMEM);/* scanner initialization & config.*/
  ts_chars(scanner,TS_FLDSEP,",\t");/* set field separators, ...        */
  ts_chars(scanner,TS_RECSEP,"\n"); /* ...and record separator          */
  ts_chars(scanner,TS_COMMENT,"#"); /* set indicator for comment lines  */
  /* blank symbols serve to distinguish elements in sets */
  if (!*fn_trn) datafile=stdin;
  else if (!(datafile=fopen(fn_trn,"r"))) error (E_OPENF,fn_trn);
  printf("\n...reading training data from file %s\n",fn_trn);
  assert(hd);
  hd->lcorr=lcorr;
  if (hdist_getfreqs(hd,scanner,datafile)!=0) error(E_PARSETBL,fn_trn);
  /* open output file and save model */
  /* empty file names are implicitly interpreted to mean stdin/stdout */
  /* Note that the single dash '-' is removed during input procssing */
  if (fn_out && !*fn_out) outfile=stdout; else outfile=fopen(fn_out,"w");
  if (outfile) {                                 /* if outputfile valid...  */
    hdist_save(hd,outfile);                      /* save the model, then... */
    if (outfile!=stdout) fclose(outfile);        /* check for special files */
    outfile=NULL;                                /* ..and close normal ones */
  }
  else {                                         /* report any failures */
    error(E_OPENF,fn_out);
  }
  cleanup();                          /* does not work according to valgrind */
  return 0;
}
