/*-----------------------------------------------------------------------------
FILE    : crsutil.c
CONTENT : auxilliary functions for command line toolbox crs..   
Author  : Frank Rgheimer
History : 2012-02-18 branched from psvmodel to provide more modular interface
                       to hdist model library; 
                     added E_STDINCNT error message and code
          2012-03-12 modified readnext_sv to be accept leading whitespace;
                     readnext_sv now properly terminates token buffers for 
                     empty set input
          2012-04-23 added E_OPTARGDSC_2 and E_READFRAME errors
-------------------------------------------------------------------------------
 Copyright 2012 Frank Rgheimer
 This program is distributed under the terms of the GNU Lesser General Public
 License.
-----------------------------------------------------------------------------*/

#include "crsutil.h"

/*-----------------------------------------------------------------------------
  error messages (from psvmodel)
-----------------------------------------------------------------------------*/
/** error message interpretation strings -- only accessed via error function */
static char *errormsg[]={
  /*E_NONE 0*/    "no error\n",
  /*E_OPENF 1*/   "failed to open file %s\n",
  /*E_EOF 2*/     "unexpected end of file %s\n",
  /*E_ATTVAL 3*/  "encountered unexpected attribute value while parsing\n",
  /*E_READFILE 4*/"error reading file\n",
  /*E_OPTION 5*/  "unknown option %s \n",
  /*E_ARGCNT 6*/  "invalid number of arguments\n",
  /*E_NOMEM 7*/   "out of memory\n",
  /*E_PARSETREE 8*/ "failure to parse tree: %s\n",
  /*E_PARSETBL 9*/"error parsing table file: %s\n",
  /*E_OPTARG 10*/ "missing option argument\n",
  /*E_BMODE 11*/  "invalid branching option argument: %s\n",
  /*E_OPTARG_M 12*/ "invalid argument for -m option: %s\n",
  /*E_LOGZERO 13*/ "cannot compute log of likelihood 0 - try \
enabling laplace correction (-l option)\n",
  /*E_LCORR 14*/  "invalid value for laplace correction\n",
  /*E_PARSE 15*/ "error while parsing file: %s\n",
  /*E_STDINCNT 16*/ "only one file can be assigned to stdin()\n",
  /*E_OPTARGDSC_2 17*/ "invalid argument for %s option: %s\n",
  /*E_READFRAME 18 */ "error while reading frame specification\n",
  /*E_UNKNOWN */ "unknown error\n"
};

/*---------------------------------------------------------------------------*/
/** generate error event and produce corresponding message
@arg  code  error code constant
@note some error messages require additional arguments; see message 
      definitions above for message format definitions; failure to supply 
      correct arguments will trigger a segmentation fault */
void error (int code, ...){
  va_list args;                        /* the variable argument stuff */
  const char* msg;                     /* the error message */
  if (code>E_UNKNOWN) code=E_UNKNOWN;
  if (code>0){
    msg=errormsg[code];
    if (!msg) msg = errormsg[E_UNKNOWN];
    fprintf(stderr,"\n%s: ",progname);
    va_start(args, code);              /* last argument before elipsis */
    vfprintf(stderr,msg,args);         /* print combined error message */
    va_end(args);                      /* finish processing varguments */
  }
  cleanup();                       /*clean up memory and exit program*/
  exit(code);
}

/*---------------------------------------------------------------------------*/

/** interpret registered measure id strings by searching a lookup table
    @arg tab  pointer to measure registration table 
    @arg name string id to search for
    @return index to associated entry in tab* or (-1) for unregistered strings
    @note this mechanism was adopted from code by C. Borgelt  */
int code (const MINFO *tab, const char *name)
{                                  /* map string to integer code */
  for (;tab->name!=NULL; tab++)    /* try names from table */
    if (strcmp(tab->name,name)==0)  
      return tab->code;            /* return code if found */
  return -1;                       /* nothing was found */
}


/*=============================================================================
   this function in shared between utilities
=============================================================================*/
/** read a set-instantiation from stream, copy it to buffer and parse it 
    individual tokens will be made available via an array of string pointers 
    to the beginning of the tokens in the read buffer -- all tokens will be 
    terminated by a \0 character 
    @arg buffer  input character buffer for line parsing 
    @arg bufsize capacity limit
    @arg tokens  array of string pointers -- starting with tokens[0] the array
         is filled with any parsed tokens (in the buffer) the element following
         the last parsed token is set to NULL to provide a guardian value
    @returns 1 if valid record is found, -1 on error, 0 if no record is left 
*/
int readnext_sv(char* buffer, int bufsize, char** tokens, int maxel,\
 TABSCAN* scanner, FILE* input){
   int d=TS_FLD;  /* for scanner state; indicate no previously read record  */
   /* variables used in parsing set descriptions */
   int elcnt,c;    /* elements already read if parsing a set; char from file */
   char *token;    /* for extracting tokens from a set                       */
   char *token2;   /* set to follow_up token when parsing strings            */
   assert(scanner&&input&&buffer&&tokens);                   /* check inputs */
   assert(maxel>=0);
   elcnt=0;
   tokens[0]=NULL;                        /* initialize as empty set */
   do {                                   /* read records from file          */
    c=getc(input);
    if (c == EOF) return 0; else ungetc(c,input);     /*test for end of file */
    /* if(tfs_skip(scanner,input)!=0) continue; *//* ignore comment lines    */
    /* found a non-comment record...                                         */
    do {                                 /* ignore all fields but the last.. */
      d = ts_next(scanner,input,buffer,bufsize);    /* ... one in the record */
    } while (d == TS_FLD);
    if (d<0) {                           /* something is wrong with the file */
       return(d);                        /* return and indicate failure      */
    } 
    /* loop to process the set of values in the current field */
    elcnt=0; token=token2=buffer;
    /* skip any leading whitespace */
    while ((*token!='\0') && (ts_istype(scanner,TS_BLANK,*token))) 
        token2=(++token); 
    if (*token=='{') token2=(++token);    /* skip opening brace if present  */
    while ((*token!='\0') && (ts_istype(scanner,TS_BLANK,*token))) 
        token2=(++token);                  /* and any whitespace after brace */
    if (*token=='}') *token='\0';          /* } terminates input immediately */                    
    for (;*token!='\0';token=token2) {   /* while there is something to read */
      while((*token2!='\0')&&(*token2!='}')\
             &&(!ts_istype(scanner,TS_BLANK,*token2)))
        token2++;                        /* seek end of first token */
      if (*token2!='\0') {
	*token2='\0';                    /* terminate token and skip blanks  */
        for (token2++;ts_istype(scanner,TS_BLANK,*token2);token2++);
      } 
      if (maxel<elcnt) return(ESV_ELCNT);/* check pointer array space  */
      tokens[elcnt]=token; 		 /* register new token         */
      elcnt++;
    }  /* nodes in the set are marked */
  } while ((d > TS_REC)||!(scanner->cnt));/* stop at end of non-empty record */
  tokens[elcnt]=NULL;                    /* invalidate following tokens ptrs */
  return 1;
}

/*===========================================================================*/
