/*-----------------------------------------------------------------------------
File    : frameconv.c
Contents: convert distribution to distribution over reference frame to a new
          reference frame generated by the same hierarchy;  
          hierarchy structure must be provided by external source;
          quantitative information for probability tree is estimated from data
Author  : Frank Rgheimer
History : 20.02.2007 file created as ptree.c
          23.02.2007 probability tree components 
          25.02.2007 added hdist interface 
          25.02.2007 switched to using attset.h for attribute management
          26.02.2007 added reference frame and valnametoids
          26.02.2007 fixed makefile
          27.02.2007 added probabilistic conversion between refernce frames
          19.08.2007 improved error and warning message mechanism
	  21.08.2007 added reading forest structure from external files
          22.08.2007 added dot-like graph language support 
          28.08.2007 switched to node based tree representation 
          31.08.2007 adapted parsing and tree construction to nodes
          04.09.2007 support for set-valued distributions added:
          08.09.2007 added branch modes for treating non-leaf nodes
          09.09.2007 moved hdist code to seperate files 
          29.01.2009 moved processing of header into main program  
          02.05.2010 renamed to frameconv.c; minor mods to source & comments
          2010-10-01 adapted to new table version 
TODO    : test mods  
          add frame conversion option based on new csvmodel
          allow import of model desceription file as alternative for direct
          induction from data
-----------------------------------------------------------------------------*/

/*---includes---*/
#include "frameconv.h"
#include "hdist.h"



/*---definitions---*/

#define DESCRIPTION "enrich tree structure with probability/possibility estimat\
es from "\
"data\n"
#define VERSION "\
version 1.4 (2010-10-01)             \
         (c) 2007-2010   Frank Rgheimer\n\
attmap, tabscan and parser  modules  \
         (c) 1996-2010 Christian Borgelt\n"
/* Thanks again to Christian as I also adopted many general tasks like argument
   processsing or error handling from his open source programs                */


/*---error codes for main program---*/
#define E_NONE  0         /*no error*/
#define E_OPENF 1         /*failed to open file */
#define E_EOF 2           /*unexpected EOF*/
#define E_ATTVAL 3        /*unknown attribute value*/
#define E_READFILE 4
#define E_OPTION 5        /* unknown option*/
#define E_ARGCNT 6        /* invalid number of arguments */
#define E_NO_MEM 7        /* out of memeory */
#define E_PARSETREE 8     /* failure to parse tree: */
#define E_PARSETBL 9      /* error parsing table file: */
#define E_OPTARG 10       /* missing option argument */
#define E_BMODE 11        /* invalid branching mode option argument: %s */
#define E_UNKNOWN 12      /*unknown error*/


/*---error messages---*/

/* error must be called with matching parameters. Failure to to so will
   usually trigger a segmentation fault                                  */
static char *errormsg[]={
  /*E_NONE 0*/    "no error\n",
  /*E_OPENF 1*/   "failed to open file %s\n",
  /*E_EOF 2*/     "unexpected end of file %s\n",
  /*E_ATTVAL 3*/  "encountered unexpected attribute value while parsing\n",
  /*E_READFILE 4*/"error reading file\n",
  /*E_OPTION 5*/  "unknown option %s \n",
  /*E_ARGCNT 6*/  "invalid number of arguments\n",
  /*E_NOMEM 7*/   "out of memory\n",
  /*E_PARSETREE 8*/ "failure to parse tree: %s\n",
  /*E_PARSETBL 9*/"error parsing table file: %s\n",
  /*E_OPTARG 10*/ "missing option argument\n",
  /*E_BMODE 11*/  "invalid branching option argument: %s\n", 
  /*E_UNKNOWN */ "unknown error\n"
};

/*---globals---*/
static char *progname;
int	    options=0;      /* for option flags, e.g. set-valued eval */
TABSCAN      *scanner=NULL;  /* scanner for table file */
SCAN        *scan=NULL;     /* general scanner */
static FILE *datafile=NULL; /* a pointer to a data file which constist of a 1 
                     line header with <space>-separated attribute names and
                     one line for each  record with <space>-separated attribute values.
		   */ 
static FILE *treefile; /* used to read tree structure */
/* format close to dotty language see example.htr */ 

typedef struct {      /* for comfortable access to option params */
  int code;           /* numeric code */
  char* name;         /* name */
} MINFO;              /* from from C.B. */


static const MINFO modemap[] = {
  {HDBM_NONE, "none"},    /* no mode selected */
  {HDBM_CONJ, "conj"},    /* conjunctive interpretation */
  {HDBM_DISJ, "disj"},    /* disjunctive interpretation */
  {HDBM_SPLIT,"split"},   /* split according to distrib. of other cases */   
  { -1, NULL}             /* invalid code (sentinel) */
};

/*-----------------------------------------------------------------------*/ 
/*---function definitions---*/
/*-----------------------------------------------------------------------*/

static int code (const MINFO *tab, const char *name)
{                                  /* map string to integer code */
  for (;tab->name!=NULL; tab++)    /* try names from table */
    if (strcmp(tab->name,name)==0)  
      return tab->code;            /* return code if found */
  return -1;                       /* nothing was found */
}



/* some error messages have arguments; see header file for individual 
   message format; failure to supply these will usually trigger a
   segmentation fault                                                    */

static void error (int code, ...){
  va_list args;                    /* the variable argument stuff */
  const char* msg;                 /* the error message */
  if (code>E_UNKNOWN) code=E_UNKNOWN;
  if (code>0){
    msg=errormsg[code];
    if (!msg) msg = errormsg[E_UNKNOWN];
    fprintf(stderr,"\n%s: ",progname);
    va_start(args, code);          /* last argument before elipsis */
    vfprintf(stderr,msg,args);     /* print combined error message */
    va_end(args);                  /* finish processing varguments */
  }
  if (scanner) ts_delete(scanner); /*clean up and exit program */
  if (scan) sc_delete(scan);
  if (treefile) fclose(treefile);
  if (datafile) fclose(datafile);                           
  exit(code);
}



#if false


/*-----------------------------------------------------------------------*/
/*    reference frame functions                                          */
/*-----------------------------------------------------------------------*/


REFFRAME* rf_create(const HDIST *hdist, const int* sel, int size){
  REFFRAME *rf;                          /* structure to return          */
  assert(hdist && sel && (size>0));      /* check function arguments     */
  if (!(rf=(REFFRAME*)malloc(sizeof(REFFRAME)))) /* reserve mem for structure*/
    return NULL;                         /* abort on failure to get mem   */
  if (!((rf->sel)=malloc(size*sizeof(int)))){    /* get mem for index map */
    free(rf);                            /* if op failed return allocated */
    return NULL;                         /* memory and indicate failure   */
  }
  memcpy(rf->sel,sel,size*sizeof(int)); /*copy reference map */
  rf->size=size;                        /*and copy size information */
  /*--check validity of frame--*/
  
  /* TODO implement check for refinement (not required for experiment) */  
  return rf;
}

void rf_delete(REFFRAME* rf) {
  assert(rf);
  if (rf->sel) free(rf->sel);
  free(rf);
}


/*-----------------------------------------------------------------------*/
void rf_show(REFFRAME *rf, HDIST *hdist)
{           /*---show leaf attribute values w.r.t. given  refernce frame-*/
  int i;                               /* a counter */
  ATT *att;

  assert(hdist);                       /*check parameter*/
  att=hdist->att;                      /* shortcut */
  printf("\nshowing base values of reference frame %p:\n",(void*)rf); 
  for (i=0;i<(rf->size);i++)
    printf(" %s ",att_valname(att,rf->sel[i]));
  printf("\n");
}
   

/*-----------------------------------------------------------------------*/
/*  auxilliary functions                                                 */
/*-----------------------------------------------------------------------*/

int  valnamestoids(ATT *att,const char **names, int *ids, int size)
{                /*---convert arrays attvalue names  to attvalue ids---*/
  int i,v;       /* counter, converted value */
  int *pi;       /* for traversing  ids */
  const char **attname; /* for traversing **names */
  
  assert (att && names && ids && (size>0));
  pi=ids;
  attname=names;                           /* traverse arrays */
  for (i=0;i<size;i++){                    /* for each element ...*/  
    *(pi++)=v=att_valid(att,*(attname++)); /* convert the attribute name*/
    if (v<0) error(E_ATTVAL);              /* but abort on invalid values*/
  } 
  return 0;
}

/*-----------------------------------------------------------------------*/
void _match(i,d){                         /* for debugging */
  fprintf(stderr,"src: %i, dst: %i\n",i,d);
}

int convdist(const HDIST *hdist, const REFFRAME *src,\
	     const double *srcdist, const REFFRAME *dst, double *destdist)
{ /*---convert distributions between frames---*/
  int i,id,didx,d; /* counter, id of attribute values, indices */ 
  double factor;   /* probability factor for general to specific conversion */
  int *todstidx;   /* map for loc indices of dst base attributes -1 for nil */
  char *matdst;    /* marker for already matched values in the dst frame */
  ATT  *att;       /* flat version of hdist attribute for managing all values */
  
  
  assert(hdist && src && srcdist &&  dst && destdist); /*check args*/
  att=hdist->att;
  todstidx=malloc((hdist_valcnt(hdist))*sizeof(int));  /* get mem for arrays */
  matdst=malloc(dst->size*sizeof(char));
  if (!(todstidx && matdst)) {
    fprintf(stderr,"\n convdst: Memory allocation failed - aborting\n");
    exit(-1);
  }
  /* (would be faster if indices in dst were sorted) */
  
  for (i=0;i<(hdist_valcnt(hdist));i++) todstidx[i]=-1; /* init array */
  for (i=dst->size;i>0;){      /*for each attribute of the destiantion frame*/
    destdist[--i]=0.0;                            /* ... clear distribution */
    matdst[i]=0;                                  /* ... clear flags */
    id=dst->sel[i];                /* get an attid from the destination frame*/
    todstidx[id]=i; /* and provide the backward reference in the full repres.*/
  } 
  for (i=0;i<(src->size);i++) {     /* for each source frame attribute value */
    for (id=src->sel[i];id>0;) /* perform specific to general search ...  */
    {                 /* ... for id of matching hatt value in dst frame */
      if ((didx=todstidx[id])>=0)
      {                  /*if matching value is found in dst frame*/
        destdist[didx]+=srcdist[i]; /* contribute probability,   */
        /*_match(i,didx);*/         /* show matches */
        matdst[didx]=1;             /* mark dst value as matched */
        break;                      /* ...and stop looking  */
      }
      id=hdist->parent[id]; /*otherwise try a more general attribute value */
    }  
    id=src->sel[i];  /* set backward references for later specific */
    todstidx[id]=i;  /* to general searches starting from dst att-values */
  }
  /* all remaining destination attributes are specialisations */
  for(d=0;d<(dst->size);d++) /*for the remaining values in the destination*/
  {             /* look for generalizations among the src-frame attributes*/
    if (matdst[d]==1) continue;    /* value has already been matched */
    id=dst->sel[d];                /* obtain true index of value     */ 
    factor=hdist->cprobs[id];      /* keep track of probability splits*/
    id=hdist->parent[id];          /* start search with parent */
    while (id>=0) {                /* while there are parents */
      if ((i=todstidx[id])>=0){/* if there is a matching value in src frame */
        /*_match(i,d);*/
        destdist[d]=factor*srcdist[i]; /*assign  partial contribution */
        break;
      }
      factor*=hdist->cprobs[id]; /* accumulate probability factor on path    */
      id=hdist->parent[id];  /* otherwise try a more general attribute value */
    }
  }
  free(todstidx);                                  /* return reserved memory */
  free(matdst);               
  return 0;                    
}
#endif 

/*-----------------------------------------------------------------------*/


/*-----------------------------------------------------------------------*/

  
/*---dummy help message---*/

static void help (void){
  fprintf(stderr,"\n");
  fprintf(stderr,"Sorry no detailed help available yet, run %s without arguments for usage\n",progname);  
}

/*---main function---*/

int  main(int argc, char* argv[]){
  int i, k=0;                      /* loop variable and argument counter */
  char       *s;                   /* for scanning options */
  char       *bmstr   = NULL;      /* string to abbreviate mode of prob. */
                                   /* distrib. to child nodes            */
  char       **optarg = NULL;      /* for redirecting  option arguments */
  char       *fn_htr  = NULL;      /* name of heirarchy tree file */
  char       *fn_tab  = NULL;      /* name of table file used for stats */
  int        bmode    = HDBM_DISJ; /* standard mode for child management */
  HDIST      *hd;                  /* hierarchical distribution */

#if false
  /* -------- code for testing -------*/
  const char *framevals[] = {"A1", "A2","A31","A32"};
  const char *framevals2[]= {"A11","A12","A13","A2","A3"};
  const char *framevals3[]= {"A11","A12","A13","A2","A31","A32"};
  int  frameids[4]; 
  int  frameids2[5]; 
  int  frameids3[6];
  REFFRAME *frame, *frame2, *frame3;   /*reference frames*/
  double sd[4]= {0.1,0.3,0.2,0.4};
  double dd[6];
  /* ---------------------------------*/
#endif

  progname=argv[0];                /* get program name for messages */
  
  /* --- print greeting message --- */
  if (argc > 1) {                  /* if there are arguments given*/   
    fprintf(stderr, "%s - %s\n", progname, DESCRIPTION);
    fprintf(stderr, VERSION); }    /* print greetings... */
  else{                            /* ...otherwise...*/
    printf("usage: %s [options] treefile infile\n", progname);
    printf("%s\n", DESCRIPTION);
    printf("%s\n", VERSION);       /* ...print usage message... */
    printf("list of supported options: \n");
    printf("-s#      set semantics of non-leaf nodes \n");
    printf("         conj  - node is replaced by set of all leaf-nodes\n"
	   "               in corresponding subtree (e.g. imprecision)\n");
    /*printf("         disj  - tuple represents case that belongs to node \n" 
      "              but none of the subnodes in the hierarchy\n");*/
    /* disjunctive has been disabled until fixed */
    printf("         split - specific assignment of case is unknown - \n" 
           "               split according to the distribution known \n"
           "               cases\n"); 
    printf("         (default: conjunctive)\n");
    return 0;                      /* ... and abort program */
  }
   
  /* --- evaluate arguments --- */ 
  for (i=1; i< argc; i++) {        /* traverse the argument vector   */
    s=argv[i];                     /* get next argument */
    if (optarg) {                  /* if at start of an option argument  */
      *optarg=s;                   /* set resp. var to reading position  */
      optarg=NULL;                 /* doubles as an indicator tnx to C.B */
      continue;                    /* read next command line argument    */
    }
    /* otherwise */
    if ((*s == '-') && *++s) {     /* for options ignore leading '-' */
      while (*s) {                 /* may be list of options */
        switch (*s++) {            /* evaluate options */
        case '?': help();                  return 0;
	case 's': optarg=&bmstr;           break;      
	default : error(E_OPTION,s-2);     return 0;
	}
        /* option arguments may follow directly or in the next cmdarg */
        if (!*s) break;            /* end of cmdarg - try next one  */
        if (optarg) {*optarg =s; optarg=NULL; break;} 
        /* identify rest of current cmdarg with the option argument */
      }
    }
    else {                         /* if standard argument */
      switch (k++) {               /* evaluate arguments acc. to position */
        case  0: fn_htr=s;      break;
        case  1: fn_tab=s;      break;
      default: error(E_ARGCNT); break; /* too many arguments */
      }
    }
  }
  if (optarg) error(E_OPTARG);     /* missing option argument */ 
  if (bmstr!=NULL) {
    bmode=code(modemap,bmstr);     /* convert -s option arg. from string */
    if (bmode<0||bmode==HDBM_NONE) error(E_BMODE,bmstr);
    /* also temporarily disabled HDBM_CONJ until fully implemented.  */
    if (bmode==HDBM_CONJ) {
      fprintf(stderr,"%s: Warning, branch mode not implemented\
 -- using default\n",progname);
      bmode=HDBM_DISJ;
    }
  }
  if (k<2) error(E_ARGCNT);        /* check required arguments */
 

  /* --- read tree structure from file --- */
  assert(fn_htr);
  scan= sc_create(fn_htr);         /* open the tree file for scanning */
  if (!scan){
    if  (!fn_htr || !*fn_htr) error(E_NO_MEM);
    else error(E_OPENF, fn_htr);
  }
  pa_init(scan);                   /* enable parser extensions */
  hd=hdist_parse(scan,bmode);      /* and build the dist tree   */
  if (!hd) error(E_PARSETREE,fn_htr);
  sc_delete(scan);                 /* close file and scanner */
  scan=NULL;
  /* --- print it to stdout --- */
  hdist_print(hd);
  
 
  /* -------- this is code for testing  ------------- */
  
  if (-1) printf("-1 is true\n"); else printf("-1 is false\n");
  if (0) printf(" 0 is true\n"); else printf(" 0 is false\n");
  if (1) printf(" 1 is true\n"); else printf(" 1 is false\n");
  if (NULL) printf(" 1 is true\n"); else printf(" NULL is false\n");
 
  
  /* force error missing an argument */
  printf("\nTesting hierarchical distribution...\n\n");
 
  assert(hd);
  assert(fn_tab);
  scanner=ts_create();
  ts_chars(scanner,TS_FLDSEP,",\t"); /* set field separators, ...         */
  ts_chars(scanner,TS_RECSEP,"\n");  /* ...and record separator           */
  ts_chars(scanner,TS_COMMENT,"#");  /* set indicator for comment lines   */
  /* blank symbols serve to distinguish elements in sets */
  if ((datafile=fopen(fn_tab,"r"))){
    printf("\ntrying again\n");
    /* while (tfs_skip(scanner, datafile)!=0); */
    /* new ts_next function automatically ignores comments */
    while (ts_next(scanner,datafile,NULL,0)==TS_FLD); /* and skip header */ 
    if (hdist_getfreqs(hd,scanner,datafile)!=0) error(E_PARSETBL,fn_tab);
    printf("\nnew counts obtained using hdist_getfreq(...)\n");
    hdist_print(hd);
    printf("\n\n");
    fclose(datafile);                    /* and close the data file          */
  } else error (E_OPENF,fn_tab);
  
   
/*   /\*---testing refernce frame functions---*\/ */
/*   printf("converting attribute values to ids...\n"); */
/*   i=valnamestoids(hd->att,framevals,frameids,4); */
/*   i=valnamestoids(hd->att,framevals2,frameids2,5); */
/*   i=valnamestoids(hd->att,framevals3,frameids3,6); */
/*   /\* for (i=0;i<4;i++) printf("%i ",frameids[i]); printf("\n"); *\/ */
/*   frame=rf_create(hd,frameids,4);       /\*make a reference frame*\/ */
/*   frame2=rf_create(hd,frameids2,5);     /\*   ...and another one *\/ */
/*   frame3=rf_create(hd,frameids3,6); */
/*   assert(frame && frame2 && frame3); */
/*   rf_show(frame,hd); */
/*   rf_show(frame2,hd); */
/*   rf_show(frame3,hd); */
/*   /\*testing frame conversion *\/ */
/*   printf("\ndistribution of frame:\n"); */
/*   for(i=0;i<frame->size;i++) printf("%.4f ",sd[i]); */
/*   printf("\nconverted from frame to frame2\n"); */
/*   convdist(hd, frame,sd, frame2, dd); */
/*   for (i=0;i<(frame2->size);i++) printf("%.4f ",dd[i]); */
/*   printf("\nconverted from frame to frame3\n"); */
/*   convdist(hd, frame,sd, frame3, dd); */
/*   for (i=0;i<(frame3->size);i++) printf("%.4f ",dd[i]); */
/*   printf("\n"); */
  ts_delete(scanner);                 /* clean up memory */ 
/*   rf_delete(frame); */
/*   rf_delete(frame2); */
/*   rf_delete(frame3); */
  hdist_delete(hd); 

  /* ------------- end of testing code ------------ */
  return 0;
}
