/*-----------------------------------------------------------------------------
FILE    : frdebug.h
CONTENTS: Simple macro definitions for debugging and rudimentary unit tests
AUTHOR  : Frank Rügheimer
NOTE    : debug mode is not thread safe (as "pseudo-local" vars may clash)
VERSION : 1.1
HISTORY :   2011-09-?? organizing standard test function into
                       this header file and added frtest macros
            2011-09-11 productive version   
            2012-02-29 added copyright statement and doxygen comments in prep.
                       of release with source package of LGPL project 
-------------------------------------------------------------------------------
Copyright 2011, 2012 Frank Rügheimer
-------------------------------------------------------------------------------
 This program may be distributed under the terms of the GNU Lesser General
 Public License                
-----------------------------------------------------------------------------*/

#ifndef INCL_FRDEBUG                    /* guard prevents multiple inclusion */
#define INCL_FRDEBUG

/** @file   frdebug.h
    @author Frank Rügheimer
    @brief  Macro library for debugging, also provides simple unit test 
            framework for C
*/

#define FRDEBUG fprintf(stderr, \
                "*** reached line %d in file %s ***\n",__LINE__,__FILE__);
#define FRDEBUGMSG(c)	 (fprintf(stderr,"*** %s ***\n",(c)))


/** unit test macros (To track failed test each test must be registered with a)
    of a named group with separate counters. Test groups are defined as local
    variables in the test function. The function frt_check(frt,cond) is then
    used to conduct a test of condition cond for test group frt. Macros 
    frt_summarize and and frt_reset respectively show and zero the test result
    counters using the stringified variable name of the test group 
    Note: use asserts for simple tests outside of group association  */
 
#if 0    
    USAGE example: 
    
    void simple_test(){
      ...
      FRTEST tst_rel;      /* define a test group */
      FRTEST tst_eql;      /* define another one  */      
      frt_reset(tst_rel);  /* initialize counters */
      frt_reset(
      ...
      /* now some mixed test */
      frt_check(tst_rel,4<5);
      frt_check(tst_eql,3=3);
      frt_check(tst_rel,!(3>3));
      frt_check(tst_eql,3=7%4);
      frt_summarize(tst_rel);
      frt_summarize(tst_eql);
    }
    
#endif
    

/** named test group structure intialized */
typedef struct _frtest{
  int testcnt;
  int failcnt;
} FRTEST;

/** reserved name for "pseudo-local variable" */

/* NOTE: we are defining local variables in within the scope of a do-while 
         block. gcc is fine with this but I am not sure  we are guaranteed to  
         get a local scope with every compiler out there                     */
/** macro-based inline pseudo functions (do while (0) used for grouping)     */ 

#define frt_check(frt,cond) do {                                            \
          FRTEST *frt_ptst=&(frt);/* we write -> emulate call by reference*/\
          if (!(cond)){fprintf(stderr,                                      \
               "Failed test \"%s\" in file %s, line %d\n",                  \
               (#cond),__FILE__,__LINE__); (frt_ptst->failcnt)++;           \
          } (frt_ptst->testcnt)++;                                          \
        } while (0)
                
#define frt_reset(frt) do {                                                 \
           FRTEST *frt_ptst=&(frt); /* call by reference enables writing */ \
           frt_ptst->failcnt=frt_ptst->testcnt=0;                           \
        } while (0)
        
#define frt_summarize(frt) do {                                             \
           FRTEST *frt_ptst=&(frt); /* must not assign structs with = */    \
           fprintf(stderr,"test group: %-20s %5d tests, %5d failed\n", \
                           (#frt),frt_ptst->testcnt,frt_ptst->failcnt);     \
        } while (0)
                                  

#endif 
