/*-----------------------------------------------------------------------------
File    : hdist.c
Contents: data structure for dealing with set-valued data with hierarchical
          attribute domain
Author  : Frank Rgheimer
History : 20.02.2007 file created as ptree.c
          23.02.2007 probability tree components 
          25.02.2007 added hdist interface 
          25.02.2007 switched to using attset.h for attribute management
          26.02.2007 added reference frame and valnametoids
          26.02.2007 fixed 32/64 bit conflict in makefile
          27.02.2007 added probabilistic conversion between reference frames
          19.08.2007 read tree structure from external files
	  21.08.2007 added reading forest structure from external files
          22.08.2007 added dot-like graph language support :
          28.08.2007 switched to node based tree representation 
          31.08.2007 adapted parsing and tree construction to new nodes
          04.09.2007 support for set-valued distributions added
          08.09.2007 added branch modes for treating non-leaf nodes
          09.09.2007 hdist.h created for hierarchical domain management
                     moded hdist-structure from ptree.c to hdist.c
          06.12.2008 made ";" line termination in tree input optional
          09.01.2009 intoduced {} set notation support from tab file
          11.01.2009 added option for Laplace correction in model induction
          13.01.2009 added simplified sv-mode (without using additional nodes)
          13.01.2009 access functions for marginal rates and coverage factors
          15.01.2009 started working on computation of sample log-likelihood
          18.01.2009 finished first version of  sample likelihood computation 
          20.01.2009 fixed bug due to assigning svfreq to the wrong nodes
          20.01.2009 removed automatic header skipping from hdist_getfreqs()
          22.01.2009 fixed bug due to misread node flag status in _logsubtree()
          22.01.2009 fixed bug due to implicit inclusion of singleton branch 
                      factor when computing subtree probabilities for set ele-
                      ments in _logsubtree() (now controlled by fromset param.)
          09.05.2009 removed some debug outputs from productive version
          04.05.2010 expanded tree parser for better graph file lang. support
          2010-08-31 worked on robust import and model saving for v2.0
          2010-09-02 finished adapting parser routine for reading optional node
                     attributes, added support for node labels
          2010-09-06 expanded hdist_print to support optional label output
          2010-09-11 implemented load/save functionality for model parameters
          2010-09-18 implemented saving routine for complete data structure 
          2010-09-19 loading routine finished -> version 2.1; mdl  
          2010-10-01 adapted to new table package version: tfscan -> tabscan,
                     attset->attset (new), parse.h->parse.h(new) 
                     v2.1-> v2.2 
          2012-02-08 fixed some typos in comments, changed documentation style
                     for doxygen 
          2012-02-13 modified comments 
          2012-02-19 updated copyright notice, synced version number in .h file
          2012-03-04 fixed bug with rare false reports of impossible events in
                     _logpsubtree when using model without Laplace correction.
                     Disabled error on (now correctly detected) impossible 
                     events as log(0.0)=-inf is valid result and is properly 
                     handled with floats that comply to IEEE standard. 
          2012-03-20 made some extensions in preparation of additional view and
                     evaluation modes
          2012-04-22 started implementing frame-dependent evaluation mode
          2012-04-26 added showframe function  
          2012-05-07 adapted query functions to handle HDNF_NOEXPAND flag
          2012-05-10 changed application of Laplace correction to svrate to 
                     prevent unassiged probability mass in refinements with
                     only one child node and large Laplace corrections
                     -> new minor version as outside behavior changes
TODO:     - add suite of unit tests,
          * fix function prototypes triggering warnings for not being 
            const-correct
          * make release as version 2.3.0 + web, also release old version 
            linked to IPMU papers
Later:
          - add -e switch to enable optional mass assignment to empty sets
            (implement as global tree property, requires modifications to 
            parser settings, 1st level normalization and query system. This 
            is  rather academic for now - all known public resources for set
            data are annotations or property listings for which the empty set 
            is not in the value range.)  
-------------------------------------------------------------------------------
 Copyright 2007-2010, 2012 Frank Rgheimer
 This program is distributed under the terms of the GNU Lesser General Public
 License.
-----------------------------------------------------------------------------*/

#include "hdist.h"

/*---definitions---*/
#define BLKSIZE 32       /* Blocksize for vectors */
#define DIGITS  16       /* resolution of floats in savefile (decimals) */

/*---hdist_warnings and errors---*/
#define HDW_NONE 0   /* no warning */
#define HDW_CNODE 1  /* could not create node %s */
#define HDW_DUPPAR 2 /* duplicate parent - ignored edge %s -> %s */
#define HDW_CEDGE 3  /* failed to create edge %s -> %s */
#define HDW_CYCLE 4  /* cycle detected - ignoring edge %s -> %s */
#define HDW_NOMEM 5  /* not enough memory */
#define HDW_MODE  6  /* unknown mode selected */
#define HDW_COUNT 7  /* could not obtain frequencies */
#define HDW_URNODE 8 /* encountered unregistered value %s*/
#define HDW_READREC 9 /* error reading record %i */ 
#define HDW_NOIMPL 10 /* mode not implemented - using default */
#define HDW_LOADPAR 11/* unrecognized parameter while loading model: %s */
#define HDW_LOADMOD 12/* unrecognized branch mode while loading model: .'%s */
#define HDW_LOAD_LCO 13 /* invalid parameter format while loading model */
#define HDW_LOADNODE 14 /* invalid node description while loading model */ 
#define HDW_UNKNOWN 15  /* unknown warning */


/*---hdist_warning message---*/

double EPSILON=1e-12;   /* Epsilon value for handling positives */
double TOLERANCE=1e-09; /* tolerance for error detection in float results */

static char *hdwmsg[]={
  /* HDW_NONE 0*/      "no warning\n",
  /* HDW_CNODE 1 */    "could not create hierarchy node:%s\n",
  /* HDW_DUPPAR 2 */   "duplicate parent - ignored edge %s -> %s\n", 
  /* HDW_CEDGE 3 */    "could not create edge %s -> %s\n",
  /* HDW_CYCLE 4 */    "cycle detected - ignoring edge %s -> %s\n",
  /* HDW_NOMEM 5 */    "out of memory\n",
  /* HDW_MODE 6  */    "unknown mode selected\n",
  /* HDW_COUNT 7 */    "could not obtain frequencies\n",
  /* HDW_URNODE 8 */   "encountered unregistered node %s - value ignored\n",
  /* HDW_READREC 9 */  "error parsing record at line %i\n",
  /* HDW_NOIMPL 10 */  "branch mode not implemented - using default\n",
  /* HDW_LOADPAR 11 */ "unrecognized parameter while loading model: %s\n",
  /* HDW_LOADMOD 12 */ "unrecognized branch mode while loading model: %s\n",
  /* HDW_LOAD_LCO 13 */ "invalid parameter format while loading model\n", 
  /* HDW_LOADNODE 14 */ "invalid node description while loading model\n",
  /* HDW_NOTINFRM 15 */ "node inaccessible in selected frame\n" 
  /* HDW_UNKNOWN 16 */  "unknown warning\n"
};

/*-----------------------------------------------------------------------------
     hdist auxilliary structures 
-----------------------------------------------------------------------------*/
/** used to report errors and inconsistencies with in instantiations */
typedef struct _inst_status{
  int err;              /**< status flag 0 for no error */
  int errind;           /**< error source indicator, e.g. nodeID */ 
} INST_STATUS;

/*-----------------------------------------------------------------------------
     hdist general-use auxilliary functions                              
-----------------------------------------------------------------------------*/

/** create directed link between two given HDIST nodes 
    @arg par ptr to node of edge source (at tail of edge)
    @arg child ptr to node at head of new edge */
void _linkedge (HD_NODE *par, HD_NODE *child){ 
  HD_NODE *p; /* to traverse child list */
  assert(par&&child&&par!=child);
  child->parent=par;
  if (!par->child) par->child=child;
  else {
    for (p=par->child;(p->link); p=p->link); /* go to last known child   */
    p->link=child;                           /* and append the new one   */
  }
  par->nchildren++;                          /* branch counter           */
}

/*-----------------------------------------------------------------------*/
/** auxillary function for depth-first tree traversion in for loops        
    @arg node a node   if (expand) -> children are searched too (standard true)           
    @int expand set expand to 0 to prevent further subtree expansion at node,
                that is, node is treated as if it was a leaf (this is useful
                generate emulate pruned trees without actually modifying the 
                underlying data structures
                Any other value enables expansion into subtrees 
    @return successor of node in depth first search */
/* Remark: This equals the rule of succession under absolute primogenture */
HD_NODE* _nextdf(HD_NODE* node,int expand){
  assert(node);
  if ((node->child!=NULL)&&expand)           /* if depth exp. enabled    */ 
    return node->child;                      /* try children first...    */
  for (;node!=NULL;node=node->parent)        /* try siblings next and    */
    if (node->link!=NULL) return node->link; /* siblings of the ancestors*/
  return NULL;                               /* ...until none are left   */ 
}


/*-----------------------------------------------------------------------*/
/*    hdist main functions                                               */
/*-----------------------------------------------------------------------*/

/** interpret and print messages for hdist-related warning code 
    @arg code message id 
    @arg ... message-specific arguments, e.g. file names 
    @return error code for passing it on to caller */
static int hdist_warn (int code, ...){
  va_list args;                    /* the variable argument stuff        */
  const char* msg;                 /* the warning message                */
  if (code>HDW_UNKNOWN) code=HDW_UNKNOWN;
  if (code>0){                     /* if there is a warning...           */
    msg=hdwmsg[code];
    if (!msg) msg = hdwmsg[HDW_UNKNOWN];
    fprintf(stderr,"\nWarning: ");
    va_start(args, code);          
    vfprintf(stderr,msg,args);     /*...print  warning  message...       */
    va_end(args);                  
  }
  return(code);                    /* ...and pass on code for..          */
}                                  /* ...further processing              */

/*-----------------------------------------------------------------------*/

/** create named hierarchical distribution structure in memory 
    @arg name name off structure for use in human readable output */ 
HDIST* hdist_create (const char *name){
  HDIST *hdist;                          /* created distribution */
  HD_NODE *root;                         /* root node of created hdist */ 
 
  assert(name && *name);                 /* check parameter */
  hdist = (HDIST*)malloc(sizeof(HDIST)); /* allocate mem for hdist */
  if (!hdist) return NULL;               /* check for successful alloc. */
  hdist->att=att_create(name, AT_NOM);   /* create internal symbolic attrib */
  if (!hdist->att){free(hdist); return NULL;}   /* check alloc poss.clean up */
  hdist->opstatus=0;
  hdist->node=NULL;                      /* initialize node reference table */
  hdist->bmode=HDBM_CONJ;                /* init branch mode flag; */
  hdist->vsz=0;                          /* init internal counters */
  hdist->lcorr=0.0;
  root=hdist->root=(HD_NODE*)malloc(sizeof(HD_NODE)); /* create root node */
  if (root) {
    root->valid=NV_NOM;                  /* not symbolic value for root */
    root->flags=HDNF_ROOT; 
    root->parent=root->child=root->svgroup=root->link=NULL;
    root->value=root->covsv=0;           /* init frequencies */
    root->nchildren=0;
    root->svrate=0.0;                    /* rate of multiple branch events */
    root->label=NULL;
  }
  else {
    att_delete(hdist->att);
    free(hdist);
    return(NULL);
  }
  return hdist;
}


/*-----------------------------------------------------------------------*/

/** hdist destructor */
void hdist_delete (HDIST *hdist){
  int i;                                 /* counter */
  HD_NODE *p;                            /* to traverse nodes */

  assert(hdist);                         /*  must exist */
  if (hdist->node) {
    for (i=hdist_valcnt(hdist);--i>=0; ){/* iterate over nodes ... */
      p=hdist->node[i];                  /* to remove the true nodes, ... */
      if (p->label) free(p->label);      /* ...any optional labels */
      if (p->svgroup) free(p->svgroup);  /* ...and any associated special */
      free(p);                           /* nodes for set-valued results */
    }
    free(hdist->node);                   /* remove data field itself */
  }
  if (hdist->att) att_delete(hdist->att);/* free the internal attribute... */
  if (hdist->root) free (hdist->root);   /* ... the root node .... */
  free(hdist);
} 


/*-----------------------------------------------------------------------*/
/** dynamically adapt the hdist node capacity at runtime
    @arg hdist hierarical distribution to be resized 
    @arg size new minimum capacity 
    @return status code (0 indicates success) */
int hdist_resize (HDIST *hdist,int size){/* --resize internal vectors-- */
  /* grow/shrink internal data structures - use -1 for shrinking        */
  int i;                               /* loop variable */
  HD_NODE **p;                         /* pointers for realloc results */
                             
  ATT *att;                            /* for shortcut to hdist->att */

  assert(hdist);
  att=hdist->att;                      /* shortcut */
  /* if (att_resize(att, size)!=0) return -1;   resize attribute vector*/
  i=hdist->vsz;                        /* get current vector sizes */
  if (size>0) {                        /* if to enlarge the vectors */
    if (i>=size) return 0;             /* vector is large enough already */
    i+= ( i>BLKSIZE ) ? (i >> 1):BLKSIZE; /* set new field size to
                                             min(BLKSIZE,1.5i)     */
    if (i>size) size=i;}
  else {                               /* size is to be reduced              */
    size=att->valcnt << 1;             /* tolerate twice the req. size       */
    if (size<BLKSIZE) size=BLKSIZE;    /* but BLKSIZE is minimum             */
    if (i<size) return 0;              /* done - if vsz small enough already */
    size=att->valcnt+(att->valcnt>>1); /* reserve for 1.5x the current #itms */
    if (size<BLKSIZE) size=BLKSIZE;    /* ensure minimum size                */
  }
  p=(HD_NODE**)realloc(hdist->node,size*sizeof(HD_NODE*)); /* realloc mem */
  if (!p) return -1;                               /* for dynamic arrays  */
  hdist->node=p;
  hdist->vsz=size;                     /* update vector size indicator    */ 
  return 0;
}

/*-----------------------------------------------------------------------*/

/** check for presence of node in the hdist structural component, if node is 
    not present yet, create the node and insert it into hdist structure       
    @arg hdist distribution to extend
    @arg name unique identifier of new node
    @return status code (0 for no error) */
int hdist_addnode (HDIST *hdist, const char *name) 
{ /* orphaned nodes will eventually assigned as direct children of root node */

  int idx, status;                 /* index of the new node, status   */
  ATT *att;                        /* shortcut for internal attribute */  
  HD_NODE *node;                   /* for the created node */
  
  assert(hdist && name && *name);  /* check inputs */
  att=hdist->att;                  /* obtain shortcut */
  if (att_valid(att,name)!=NV_NOM) /* if the node is already listed */
    return 0;                      /* ...then we are finished,  */
  if (hdist_resize(hdist,att->valcnt+1) != 0)  /* reserve  memory ...    */ 
    return (hdist_warn(HDW_CNODE)); /* ...generate warning if failed     */
  node=(HD_NODE*) malloc(sizeof(HD_NODE)); 
  if (node) {
    status=att_valadd(att,name,NULL);/* add name to internal attribute...*/
    if (status<0) {
      free(node);
      return (hdist_warn(HDW_CNODE)); /* failed at attrib.lvl  */
    }
    idx=att->inst.i;               /* get value index ...  */
    hdist->node[idx]=node;         /* ...and register node */
    node->valid=idx;               /* and store associated symbol index */
    node->child=node->svgroup=node->link=NULL;
    node->value=node->covsv=0;     /* init node fields */
    node->flags=node->nchildren=0;
    node->parent=NULL;             /* ...and flag node as orphaned*/
    node->svrate=0.0;              /* local rate of multiple branch events */
    node->label=NULL;              /* no label by default */
    /* orphans are later reassigned with root as default parent         */
    /* consult tree construction methods to see how this is addressed   */
  }
  else return (hdist_warn(HDW_CNODE));
  return 0;
}
/*-----------------------------------------------------------------------*/
 
/** add edge to the structural component of hierarchical distribution model;
    tree structure is enforced
    -- orphaned nodes are directly linked to root node 
    @arg hdist distribution to extend
    @arg name unique identifier of new node
    @return status code (0 for no error) */
int hdist_addedge (HDIST *hdist, const char *pname, const char *cname)
{ /* add a directed edge from node designated pname -> cname. Only one   */
  /* parent (the rootward node) is permitted for each child. Any edges   */
  /* that violate this constraint or close a directed cycle are rejected */
  int cidx, pidx;                  /* parent (near root) and child index */ 
  int status=0;                    /* for status code of called routines */
  ATT *att;                        /* for shortcut to internal attribute */
  HD_NODE *ppn;                    /* pointer to parent node             */
  HD_NODE *pcn;                    /* pointer to child node              */
  HD_NODE *p;                      /* for traversing ancestors           */

  assert (hdist && cname && *cname && pname && *pname);/* check args     */
  att=hdist->att;
  cidx=att_valid(att,cname);       /* try to obtain indices if node are  */
  pidx=att_valid(att,pname);       /* ...already in the structure        */
  pcn=(cidx==NV_NOM ? NULL : hdist->node[cidx]); /* get  pointers of any */
  ppn=(pidx==NV_NOM ? NULL : hdist->node[pidx]); /* registered nodes     */

  if (pcn!=NULL){                  /* if child node is already known     */
    if ((pcn->parent!=NULL)&&(ppn!=pcn->parent))  /* check for duplicate */
      return hdist_warn(HDW_DUPPAR, pname, cname);  /* ... parent asgnmt */
    for (p=ppn; p!=NULL ;p=p->parent) {   /* check directed for cylces   */
      if (p==pcn) return (hdist_warn(HDW_CYCLE,pname,cname));
    }
  }
  /* if everything is fine so far - add any implicit nodes given...      */
  if (!ppn){
    status=hdist_addnode(hdist,pname);
    pidx=att_valid(att,pname);          /* ...and recover their indices  */
    /* may be speed up a bit by using pidx=att->inst.i instead           */
    ppn=(pidx==NV_NOM ? NULL : hdist->node[pidx]);
  }
  if (!pcn){
    status|=hdist_addnode(hdist,cname);
    cidx=att_valid(att,cname);
    /* using pidx=att->inst.i instead might be faster though             */
    pcn=(cidx==NV_NOM ? NULL : hdist->node[cidx]); 
  }
  if (status) return hdist_warn(HDW_CEDGE,pname,cname); /* nodes missing */
  _linkedge(ppn,pcn); 
  return 0;
}

/*------------------------------------------------------------------------*/

/** get the number of registered nodes 
    @arg hdist hierarchical distribution model                            */
int hdist_valcnt(const HDIST *hdist){
  assert(hdist); return (hdist->att)->valcnt;
}

/*------------------------------------------------------------------------*/

int hdist_parseintern(HDIST *hd, SCAN *scan)
{ /* auxilliary parser function only to be used by hdist_parse            */
  /* reads a directed graph in simplified dotty language                  */
  /* scanner should take care of comments                                 */

  /* check if safe from buffer overrun (token lenght limited by buf size) */
  char pnbuffer[SC_BUFSIZE]; /* buffer for parent name when reading edges */ 
  char ptbuffer[SC_BUFSIZE]; /* buffer for node parameter type IDENTIFIER */
  int  nid;                  /* used to store id of current node          */ 
  assert(hd && scan);  
  /* check if first token is 'digraph'                                    */
  GET_TOK();                       /* read first token                    */
  if ((sc_token(scan) == T_ID)     /* consume optional strict             */
      && (strcmp(sc_value(scan),"strict") ==0)) GET_TOK(); 
  if ((sc_token(scan) != T_ID)
      || (strcmp(sc_value(scan),"digraph") !=0)) ERR_STR("digraph");       
  GET_TOK();                                         /* consume 'digraph' */
  if (sc_token(scan) !=T_ID) PERROR(E_ATTEXP);
  att_rename(hd->att,sc_value(scan));          /* set read attribute name */
  GET_TOK();                                    /* consume attribute name */
  GET_CHR('{');                                            /* consume '{' */
                                                  /* parse edge sequences */
  while ((sc_token(scan)==T_ID)||(sc_token(scan)==T_NUM)) { 
    if ((strcmp(sc_value(scan),"node")==0) 
         || strcmp(sc_value(scan),"edge")==0){
      GET_TOK(); 
      if (sc_token(scan)=='[') {   /* consume any default node/edge attribs */
        do GET_TOK(); while (sc_token(scan)!=']'); 
      }
      GET_TOK();
      if (sc_token(scan)==';') 
        GET_TOK();                 /* consume optional ";" */
      continue;                    /* ignore special cmds to set defaults */
    }  
    hdist_addnode(hd,sc_value(scan));       /* insert node if necessary   */  
    nid=hdist_nodeid(hd,sc_value(scan));    /* remember current node id   */
    strncpy(pnbuffer,sc_value(scan),sc_len(scan)+1); /* safe name to buf  */
    GET_TOK();
    if (sc_token(scan)=='['){               /* go through node attributes */
      GET_TOK();                            /* moved out of for loop      */
      for (;sc_token(scan)!=']';){
        /* this requires testing; label attribute not evaluated yet       */ 
        /* check call of error function!                                  */
        if (sc_token(scan)!=T_ID) PERROR(E_PAREXP); /* read attribute name   */
        strncpy(ptbuffer,sc_value(scan),sc_len(scan)+1);/* note param string */
        GET_TOK();                                 /* consume variable id    */
        GET_CHR('=');                              /* consume equantion sign */
        if (strcmp(ptbuffer,"label")==0) {
          if (sc_token(scan)!=T_ID) PERROR(E_STREXP); /* check label type... */
          hdist_setnodelabel(hd,nid,sc_value(scan)); /* & copy to node field */
        }
        else if (!((sc_token(scan)==T_ID)||(sc_token(scan)==T_NUM))){
          PERROR(E_VALEXP);
        }
        GET_TOK();                               /* consume value            */
        if (sc_token(scan)==',')
          GET_TOK();                             /* read opt. separator      */
      } /* for */
      GET_CHR(']');                              /* consume ']'              */
    }   /* if in [] block */
    while (sc_token(scan)==T_RGT) {              /* parse edges              */
      GET_TOK();                                 /* consume "->"             */
      if (!((sc_token(scan)==T_ID)||(sc_token(scan)==T_NUM)))
        PERROR(E_VALEXP);
      hdist_addedge(hd,pnbuffer,sc_value(scan));
      strncpy(pnbuffer,sc_value(scan),sc_len(scan)+1);          /* save name */
      GET_TOK();                                 /* consume value token      */
    }
    if (sc_token(scan)==';') 
      GET_TOK();                                 /* consume optional ";"     */ 
  }
  GET_CHR('}');                                  /* consume final '}'        */
  return 0;                                      /* parsing successful       */
}

/*---------------------------------------------------------------------------*/

/** parse a graph structure in simplified dot-language and create a new
    hierarchical distribution model according to the structure specified 
    therein; any orphaned nodes are integrated as direct children of the 
    root node
    @arg scan file scanner pseudo-object from C. Borgelts util package
         preconfigured for reading the structure input
    @arg bmode desired branch mode flag for hdist structuere (see hdist.h)
    @return pointer to created hdist structure, NULL on failure             */
HDIST* hdist_parse(SCAN *scan,int bmode)
{
  HDIST *hd;                       /* for the result */
  int i;                           /* counter */
  HD_NODE *n;                      /* for traversing nodes */
  
  assert(scan);                    /* check argument */
  /* uses a downgraded dot-like  graph language see example.htr */
  hd=hdist_create("unnamed");      /* will be renamed later on */
  if (!hd) {hdist_warn(HDW_NOMEM); return (NULL);}
  hd->bmode=bmode;                 /* set branch interpretation */
  if (hdist_parseintern(hd,scan)!=0) {           /* run parsing */
    hdist_delete(hd);              /* on error cleanup and ...*/
    return NULL;                   /* signal failure */
  }
  /* register any orphaned nodes as children of root */
  /*for(i=hdist_valcnt(hd);--i>=0;) {  */
  for(i=0;i<hdist_valcnt(hd);i++){
    n=hd->node[i];
    if (!n->parent) _linkedge(hd->root,n);
  }
  /* originally set-valued outcomes and "other"and were represented using 
     separate nodes -- this has now been replaced with a more efficient
     representation via data fields in the node records. If reverting to
     explicit node representations this function requires code for creating 
       special nodes in HDBM_DISJ mode and
       hidden nodes fo set-valued set-valued event prob  */ 
  return hd;
}

/*-----------------------------------------------------------------------*/

/** low-level function for printing text represenation of 
    subtrees of hdist model; retains compatibility with older output formats
    based on representation of multi-values refinements via separate nodes 
    @arg hd hierarchical distribution from which print
    @arg node print root node of subtree to be printed 
    @arg level positive integer indcating desired level of indentation
         (helpful when producing human readable output) */
void _printsubtree(HDIST *hd, HD_NODE *node, int level)
{ /* auxilliary function for recursive tree printing */
  int i; 
  const char *label;
  const char *optlabel;
  const char *openopt;
  const char *closeopt;
  char mark;
  HD_NODE *p; /* for traversing children */

  assert(hd&&node);
  for (i=0;i<level;i++) printf("\t");
  label="<?>"; optlabel=openopt=closeopt="";
  if (node->label) { 
    optlabel=node->label;
    openopt=" ["; closeopt="]";
  }
  if (node->flags&HDNF_ROOT) label="<root>";
  else if (node->flags&HDNF_SV) label="<set-valued>";
  else if (node->valid!=NV_NOM) label=att_valname(hd->att,node->valid);
  mark=(node->flags&HDNF_MARK)?'!':' ';
  printf("%s%s%s%s (%f; %f) %i %f %f %f %c \n",\
    label,openopt,optlabel,closeopt,node->value,node->covsv,node->nchildren,\
    node->svrate,hdist_nodefreq(hd,node),hdist_nodecov(hd,node),mark);
  /* recursion for any children */
  for (p=node->child;p;p=p->link) _printsubtree(hd,p,level+1);
}

/*-----------------------------------------------------------------------*/
  
/** show text representation of hierarchical distribution 
    @arg hd hierarchical distribution to be shown */
void hdist_print(HDIST* hd){
  assert(hd);
  printf("[%s] \n\n",att_name(hd->att));
  _printsubtree(hd,hd->root,0);  
}

/*-----------------------------------------------------------------------*/

/** internally used by to hdist_getfreqs for updating internal counters
    increases frequency/coverage counters for previously marked nodes 
*/
void _updatecounters(HDIST *hd,double *afreq,double *acov,
                               double *asvfreq, double *pasvtop){
  HD_NODE *p;    /* parent of current branch */
  HD_NODE *c1;   /* 1st marked child of current branch */
  HD_NODE *c;    /* for traversing children */
  double *counter;/*pointer to array occurence is attributed to.Initially*/
                 /* this is afreq. If more than one child is marked the  */
                 /* asvfreq for the associated sv-node is increased in-  */
                 /* stead and counter is switched to acov. Count of first*/
                 /* element is delayed to await descision */
  int marked;    /* to save marker state of parent node */
    
  assert(hd&&afreq&&acov&&asvfreq);                       /*check inputs */
  /* traverse all branches with marked parents */
  for (p=hd->root;p!=NULL;p=_nextdf(p,marked)) {
    /* traverse a branch */
    c1=NULL;                                        /* reset first child */
    counter=afreq;                     /* initially to count frequencies */
    for (c=p->child;c!=NULL;c=c->link){    /* check children for marking */
      if (c->flags & HDNF_MARK){
	if (c1==NULL) c1=c;        /* encountered 1st marked child       */
        else {                     /* more than 1 marked child exist     */
          counter=acov;            /* ensure counter is redirected       */
          counter[c->valid]++;     /* increase coverage factor for child */
        }
      }  
    } /*end for */
    if (c1!=NULL) counter[c1->valid]++;  /* inc frq/cov of 1st child too */
    if (counter==acov) {                 /* if set-valued ...            */
      if (p->flags & HDNF_ROOT)(*pasvtop)++; /*inc top-lvl set-v counter */  
      else asvfreq[p->valid]++;          /*or inc conventional sv counter*/
    }
    marked=p->flags&HDNF_MARK;     /* save marker flag for traversion    */
    p->flags&=~HDNF_MARK;          /* reset marker flag for processed node */
  } /* end node traversion */   
}

/*-----------------------------------------------------------------------*/

/** mark a node and all its predecessors in the hierarchy - used by getfreqs 
    to keep track of instantiation/coverage events 
    @arg node ptr starting node (most specific node to be marked)        */
void _markpathto(HD_NODE *node){  /* mark the path from root to a given node*/
  HD_NODE *p;
  assert(node);
  for (p=node;p!=NULL;p=p->parent){
    p->flags|=HDNF_MARK; 
  }
}


/*-----------------------------------------------------------------------*/
/**  auxilliary function used to compute a subtrees contribution to 
     log-likelihood of a sample    
     @arg  node   ptr of root node of selected subtree 
     @arg  errind set to nozero value to indicate error replace by struct?
     @arg  fromset flag indicating subtree was reached as part of a 
           set instantiation
     @return additive constributuion to log-likelihood of current instantiation
     @note: Use of this function requires that the current instantiation has
            been converted to node markings; interpretation frames are encoded
            via expansion flags. 
 */
double _loglsubtree(const HD_NODE* node, INST_STATUS* status , 
                                                      const char flags){
  int i,k,markcnt;          /* counter for number of marked children */
  double res;               /* intermediate res (several uses) */
  double renorm, psmpl;     /* set inst. renrm. factor, set probability */
  double logvalue;          /* log-likelihood value (intermediate result) */
  double logpsubs;          /* log-likelihood of marked subset acc. to cov. */
  double cldrate;           /* probability mass attributed to sglt. children */
  double svrateeff;         /* effective probability mass assigned to set re-
                   finements: in older versions of the distribution model 
                   probability mass could be assigned to the multi-valued node
                   refinement class even if no such refinement existed 
                   (degenerate branch with one child only) - this value allows 
                   to apply a correction based on local topology without 
                   losing compatibility with saved model files from older
                   versions */ 
  char fromset;             /* indicates subtree was reached as part of a set-
                               instantiation */ 
  HD_NODE* pc;              /* to traverse children */
  HD_NODE* pmc;             /* last marked child encountered */
  assert(node && status);
  assert(node->flags & HDNF_MARK);   /* can be called for marked nodes only */ 
  cldrate=logpsubs=0.0;  
  markcnt=0; pmc=NULL;                      /* init counters/accumulators */
  fromset=(flags&HDQF_FROMSET)?1:0;  /* extract legacy fromset argument  */ 
  if (!(node->flags&HDNF_NOEXPAND)) {
    /* iterate over children, aggregate any singleton refinement probabilities 
     and count the marked children encountered, also pre-compute probability of
     a potential set instantiation corresponding to marked pattern     */
    for (pc=node->child;pc;pc=pc->link){      /* traverse children */
      cldrate+=pc->value;                     /* cumulative branch prob. */
      /* also compute probability of occurring for the set instantiation */
      if (pc->flags&HDNF_MARK) {
        if (pc->covsv<0.0)        {/* encountered invalid coverage rate value*/
          status->err=1; status->errind=pc->valid; /* set error information */
          return(DBL_MIN);       
        } 
        logpsubs+=log(pc->covsv);       /* use p-factor of element selection */
        /* for impossible events this yields -inf                            */
        markcnt++;                      /* increase counter for marked nodes */
        pmc=pc;                         /* remember last marked child */
      }
      else {
        if (pc->covsv>1.0) { /* invalid value encountered */
          status->err=1; status->errind=pc->valid; /* set error state */
          return(DBL_MIN);                         /* and exit prematurely */
        }  
        logpsubs+=log(1-(pc->covsv));   /* use prob factor for elem. absence */
        /* again if absence is impossible acc. to model this yields -inf  */ 
      }
    } /* markcnt and logpsubs known, any ptr to singleton refs stored in pmc */
    /* consider node value, unless current instantiation is a set */
  }
  if (node->value>=0.0) logvalue=fromset?0.0:log(node->value); 
  else {                 /* illegal nodevalue */
    status->err=1, status->errind=node->valid;
    return(DBL_MIN);
  }
  /* printf("_loglsubtree called...  ");
  printf("nodeid %i, cldrate %.3f, svr %.3f, mc %i, psubs %.3f\n",
    node->valid,cldrate ,node->svrate, markcnt, exp(logpsubs)); */
  /* check whether multi-valued refinements are possible to determine effective
     mass towards set-valued events */
     svrateeff=(node->nchildren>1)?node->svrate:0.0;
  /* check number of marked children to select method for computing  
     likelihood of observation                               */
  switch (markcnt) {
    case  0: res=(1.0-cldrate-svrateeff);  /* no children marked */
             if (res+TOLERANCE<0.0) {         /* check for inconsistent params */
               status->err=1; status->errind=node->valid; 
               return(DBL_MIN);
             }
             if (res<0.0) res=0.0;          /* counter numerical imprecision*/
             return(logvalue+log(res));     /* log of portion of prob.mass  */
             break;                         /* not assigned to refinements  */
    case  1: return(logvalue+_loglsubtree(pmc,status,0)); /* use chain rule */
             break;        /* to expand singleton refinements down the tree */
    default: if (svrateeff+TOLERANCE<0.0) { /* true set refinement */
               status->err=1; status->errind=node->valid; 
               return(DBL_MIN);
             }
             /* compute renormalization factor of logpsubs for outcomes 
                with empty set or singletons */
             renorm=0;
             for (i=(-1);i<markcnt;i++){     /* get mass of <2 element sets */ 
               for(psmpl=1.0,k=0,pc=node->child;pc;pc=pc->link){
                 /* compute portion of mass attributed to sets with all but  
                    the i-th marked node (1st run generated empty set)      */
                 psmpl*=((pc->flags|HDNF_MARK) && (k==i))?\
                    pc->covsv:(1-pc->covsv); 
                 if (pc->flags|HDNF_MARK) k++;
               }
               renorm+=psmpl;
             }
             /* printf("renormsv: %.3f\n",renorm); */
             /* integrate selection of set-values result and selection of 
                particular observed selection pattern into likelihood */
             res=logvalue+log(node->svrate)+logpsubs-log(1-renorm);
             /* printf("before recursion: %.3f\n",exp(res)); */
             /* recursion into subtrees of marked children */
             for (pc=node->child;pc&& !(status->err); pc=pc->link){
               if (!(pc->flags&HDNF_MARK)) continue; /* skip unmarked nodes */
               /* printf("\t psubtree: %.3f\n",exp(_loglsubtree(pc,errind,1)));*/
               res+=_loglsubtree(pc,status,1);        
               /* To prevent factoring in factors in single branch probability
                  _logsubtree must be informed that we are in set mode, i.e., 
                  we want to considery consider lower level branching only 
                  (flag passed as 3rd parameter serves this purpose)*/ 
             }
             /* printf("and after %.3f\n",exp(res)); */
             return(res);   
             break;
  }
}

/*--------------------------------------------------------------------------*/

/** compute predicted log likelihood of a set-instantiation 
    @arg hd hierarchical distribution model 
    @arg array with strings representing the elements of the set-instantiation
    @arg maxel number of elements in the instantiation
    @arg failpos string pointer that will be set to first error token if 
         parsing fails -- used for error messages
    @arg flags evaluation mode flags (see hdist.
    @note: While this function will report parsing errors and often recover
           from them, illegal inputs will currently not always be reported. 
           For instance,  having both a node and its expansions in the same 
           instantiations will simply drop the extra intermediate node during 
           input conversion converting it the instantiation to a valid one.
    @note  target frame is configured via expansion flags, default: no fixed 
           frame -- if a  target any specification of a refinement of the 
           target frame will be coarsened to match the target frame.
    @todo: Write simple tests to report incorrect instantiations or treat them
           as error state to avoid confusion.
*/
double hdist_setlogl(HDIST* hd, const char **elements, int maxel, 
                                const char **failpos, const int flags){
  /* compute the likelihood of an observed set of values for current frame */
  HD_NODE* p;
  int code;
  INST_STATUS status;                          /* error indicator structure  */
  const char** elname;                       
  double  res;                                 
  
  assert(hd&&elements&&failpos);
  *failpos= NULL;                      /* no error so far            */
  /* read up to maxel elements - stop when null is encounterd        */
  for (elname=elements;elname<(elements+maxel)&&(*elname!=NULL);elname++){
#ifndef NDEBUG
    /* printf("%s ",*elname); */
#endif
    code=att_valid(hd->att,*elname);    /* decode elements...        */
    if (code<0) {
      if (*elname && **elname) {        /* if real error symbol present */
         *failpos=*elname;              /* set failure indicator to it  */
      }                                 
    }
    else 
      _markpathto(hd->node[code]);      /* mark path to observed node */
  }
  if (hd->root) hd->root->flags|=HDNF_MARK; /* required for mt set */
  /* hdist_print(hd); */ 
  /* now recursively compute the likelihood for the marked instantiation */
  status.err=0;
  res=(hd->root)?_loglsubtree(hd->root,&status,0):(0.0);
  for (p=hd->root;p!=NULL;p=_nextdf(p,1))   
    p->flags&=~HDNF_MARK;              /* clear node marks */
  if (status.err) *failpos="";         /* unknown error or zero prob. sample */
  return res;
}

/*-----------------------------------------------------------------------*/

void _markchildren(HD_NODE *node){  /* recursively mark children of node */
  HD_NODE *p;
  if (node==NULL) return;
  for (p=node->child;p!=NULL;p=p->link){
    p->flags|=HDNF_MARK;
    _markchildren(p);
  }
}

/*-----------------------------------------------------------------------*/

/** Core function for parsing instantiation table and inducing distribution
    parameters from the data therein -- only last column of an input table 
    is read; computes conditional probabilities of fine-grained attribute 
    values given coarse ones
    @note parsing has been simplified due to new table scanner -- trimming 
    strings and comment removal were dropped as they are now handled by the 
    scanner
    @arg  hdist  hierarchical distribution                  
    @arg  scanner table file scanner 
    @arg  datafile file handle for reading table data
    @returns status/error code (0 means no error) */
int hdist_getfreqs(HDIST *hdist, TABSCAN *scanner, FILE* datafile)
{                 /*---read data file and compute branch probabilities---*/
  int d,i,k;                             /* type of delimiter, counters     */
  int l;                                 /* ID of a class label             */ 
  HD_NODE *p;                            /* for traversing nodes            */
  HD_NODE *p2;                           /* for buffering parent            */
  double pcnt,psvc;                      /* count/sv count for parent - used 
                                            as normalization factors        */
  int valcnt;                            /* number of attribute values      */
  ATT *att;                              /* for shortcut altn. call directly*/
  double *data;   /* data block (used for counter arrays) */
  double *afreq;  /* counters for outcomes attributed to single reg. nodes  */
  double *acov;   /* for set-outcomes involving a registered node           */
  double *asvfreq;/* for set-outcomes attributed to nodes sv_child          */
  double acnt=0;  /* total number of observations;                          */
  double asvtop=0;/* counter for top level set valued events (not in array) */
  double lcorr;   /* shortcut to hdist->lcorr */
  int elcnt;   /* elements already read if parsing a set                    */ 
  char *token; /* for extracting tokens from a set                          */
  char *token2;/* set to follow_up token when parsing strings               */
  /* internal buffer scan->buf is limited to 255 symbols, but listed sets   */
  /* are sometimes larger. Use larger local buffer instead                  */ 
  char buffer[2048];/* create a large reading buffer on the stack for long  */
  const int bufsize=2048; 
   
  assert(hdist && scanner && datafile);
  lcorr=hdist->lcorr;                     /* get laplace correction shortcut */ 
  att=hdist->att;                         /* get shortcut to attribute    */
  /*while (tfs_skip(scanner, datafile)!=0);*//* remove lines with comments   */
  /*while (tfs_getfld(scanner,datafile,NULL,0)==TS_FLD);*/ /* ignore  header */ 
  valcnt=hdist->att->valcnt;              /* get number of values         */     
  /* dynamic counter arrays */
  afreq=data=malloc(3*valcnt*sizeof(double));      /* get required memory */
  if (!data) return hdist_warn(HDW_COUNT); 
  for (i=3*valcnt;--i>=0;) data[i]=0;    /* initialize counters    */
  acov=data+valcnt;                      /* and set array pointers */
  asvfreq=acov+valcnt;
  do {                                   /* read records from file          */
                                         /* get all field in a record ...   */
    while (TS_FLD== (d = ts_next(scanner,datafile,buffer,bufsize))){
      if (d<0) {                         /* something is wrong with the file*/
        free(data);                      /* clean up the heap structures    */
        return(hdist_warn(HDW_READREC,scanner->reccnt+1)); /* warn and exit */
      }                                  /* end (exit with a warning)       */
    }                                    /* ..until a whole record was read */
    /* count number of values and handle elements in sets (blank seperated) */
    elcnt=0; token=token2=buffer; 
    if (*token=='{') token2=(++token);   /* skip leading '{' if present */
    for (;*token!='\0';token=token2) {   /* while there is something to read*/
      while((*token2!='\0')&&(*token2!='}')&&(!ts_istype(scanner,TS_BLANK,\
*token2)))
        token2++;                        /* seek end of first token */
      if (*token2!='\0') {
	*token2='\0';                    /* terminate token and skip blanks */
        for (token2++;ts_istype(scanner,TS_BLANK,*token2);token2++);
      }
      l=att_valid(att,token);            /* determine its index code value  */ 
      if ((l<0)||(hdist->node[l]==NULL)) /* encountered unregistered value  */ 
          hdist_warn(HDW_URNODE,buffer); 
      else {                             /* mark all nodes involved in event */ 
        if (hdist->bmode==HDBM_CONJ) {   /* ... in "conjunctive/imprecision" */
          _markchildren(hdist->node[l]); /* interpret. include all children  */
	}
	for (p=hdist->node[l];p!=NULL&&!(p->flags&HDNF_MARK);p=p->parent)
	  p->flags|=HDNF_MARK;           /* mark node & any unmarked parents */     
        elcnt++;
      }
    }  /* all nodes in the multi-label instantiation S_H on hierarchy marked */
    switch (elcnt) {
    case 0:  /* no valid value was read */        break;
      /* empty set is not a valid outcome here -- so we do not include it in 
         the reference set either -- adapt parser and increase acnt here   
         to admit empty sets as valid outcomes alternatively use extra node 
         between root and it's children with competing parent assignments 
         to emulate empty set results without modifying the program itself   */ 
    default: 
      _updatecounters(hdist,afreq,acov,asvfreq,&asvtop);
      acnt++;                                     break;          
    }
	     
  } while (d > TS_EOF);                 /* stop when eof is reached          */

  /*---show internal counters in debug mode---*/
  #ifndef NDEBUG
  fprintf(stderr,"node\tafreq    acov      asvfreq\n");
  for (i=0;i<valcnt;i++){
    fprintf(stderr,"%s:\t%6.0f  %6.0f  %6.0f\n",
	    att_valname(hdist->att,i),afreq[i],acov[i],asvfreq[i]);
  }
  fprintf(stderr,"total\t%6.0f\n\n",acnt);
  #endif
  /* ---end debug output ---*/
  /*---update branch distributions---*/
  /* set valued outcome groups do not receive a laplace correction unless they
     represent at least one valid (not singleton or empty set) outcome */ 
  for (p=hdist->root;p!=NULL;p=_nextdf(p,1)){  /* traverse nodes */
    if (p->child==NULL) continue;       /* no distribution to update*/
    /* compute divisors */             
    switch (hdist->bmode) {
    case HDBM_SPLIT:                   /* compute normalized dist  */
      /* non-leaf nodes are interpreted as indicating uncertainty w.r.t.
      which child node was realized, probability mass is split proportionally 
      to children and sv-class according to distrib estimated from other child
      nodes - because the [unrefined] outcome is excluded, no Laplace 
      correction term is applied for it.*/
      pcnt=psvc=(p->flags&HDNF_ROOT)?asvtop:asvfreq[p->valid]; 
      psvc+=(2.0*lcorr);    /* norm. factor for coverage rates (binary dist) */     
      for(p2=p->child;p2!=NULL;p2=p2->link) 
        pcnt+=(lcorr+afreq[p2->valid]);     /* ref.set = children only*/
      if (p->nchildren>=2) pcnt+=lcorr;     /* +1 corr for valid sv-classes */
      break;   
    case HDBM_CONJ:                   
       /* non-leaf nodes interpreted as shortcut for a simultaneous realization
          of all possible outcomes (e.g. when used as a constraint to possible 
          states); free probability mass is credited to braches 
        * no postprocessing required  
        * implemented but no automated testing facilities yet - add them with 
          demos                                                   
        * continue with default   */
       /* hdist_warn(HDW_NOIMPL); */
    case HDBM_DISJ:                   
       /* non-leaf nodes are considered as experimental outcomes in their own 
          right; thus nodes may retain a portion of their probability mass that
          is not fixed to child nodes/subsets 
        * uses non-normalized divisor (same as default) */
    default:                           /* non-normalizing divisors */
       if (p->flags&HDNF_ROOT) {
         pcnt=acnt+(1+p->nchildren)*lcorr; /* compute laplace corrected... */
         psvc=asvtop+2.0*lcorr;            /* normalization factors */
       }
       else { 
         k=p->valid;                   
         pcnt=afreq[k]+acov[k]+(1+p->nchildren)*lcorr;
         psvc=asvfreq[k]+2.0*lcorr;
       }
       break;                          /* ref. set = parent+children */
       if (p->nchildren>=2) pcnt+=1;   /* add laplace corr. for set ref. evnt*/
    }
    for (p2=p->child;p2!=NULL;p2=p2->link){   /* update children */
      k=p2->valid;                            /* get index of child node */
      p2->value=((pcnt>0.0)?((double)afreq[k]+lcorr)/pcnt:0.0);
      p2->covsv=((psvc>0.0)?((double)acov[k]+lcorr)/psvc:0.0); 
      /*previous formulae without laplace correction */ 
      /*p2->value=((pcnt>0)?(afreq[k]/(double)pcnt):0);*//* upd. brnch prob. */
      /*p2->covsv=((psvc>0)?(acov[k]/(double)psvc):0); *//* upd. cov-factors */
    } 
    /* update svrate for node */
    /* 2 lcorr have been added to psvc for using it as a normalization factor 
       for the binary coverage distributions  modify that value again to 
       account for the relative sv rate (do not increase svrate if there are 
       fewer than two children  - no set refinmnts are possible in that case)*/
    p->svrate=((pcnt>0.0)&&(p->nchildren>=2))?(psvc-lcorr)/pcnt:0.0;
  }
  hdist->root->value=1;
  free(data);
  return 0;
}

/*--------------------------------------------------------------------------*/

/** apply chain rule of probability to reconstruct marginal frequency of 
   symbol represented by individual node. This function does not consider 
   frames so the HDNF_NOEXPAND flag is ignored!  
   @arg hdist ptr to hierarchical distribution to query
   @arg p     ptr to node for query symbol
   @return    singelton rate estimate for symbol */
double hdist_nodefreq(HDIST* hdist, HD_NODE* p){
   double res=1.0;
   HD_NODE* pc;
   assert(hdist && p);
   /*if (!(p->flags|HDNF_NOEXPAND)){ */ /* correct for prob.-loss 2 children */
     for (pc=p->child;pc;pc=pc->link) res-=pc->value;
     res-=p->svrate;
   /*}*/
   while (p) /*(!(p->flags&HDNF_ROOT))*/{
      res*=p->value;
      p=p->parent;
   }
   return res;
}

/*--------------------------------------------------------------------------*/

/** as hdist_nodefreq(), but consider contribution from set outcomes too -- 
   low level access does not take frames into account, so the HDNF_NOEXPAND 
   flag is ignored!  
   @arg hdist ptr to hierarchical distribution to query
   @arg p     ptr to node for query symbol
   @return    estim. occurence rate for symbol */
double hdist_nodecov(HDIST* hdist, HD_NODE* p){
   double res=1.0;
   HD_NODE* pc;
   assert(hdist && p );
   /*if (!(p->flags|HDNF_NOEXPAND)){ *//* if expanded - don't count prob 2 cld */
     for (pc=p->child;pc;pc=pc->link) res-=pc->value;
     res-=p->svrate;
   /*}*/
   while (!(p->flags&HDNF_ROOT)){
     res*=(p->value+p->covsv*p->parent->svrate); 
     p=p->parent;
   }
   return res;
}


/*-----------------------------------------------------------------------*/
/*  node access functions                                                */
/*-----------------------------------------------------------------------*/

/** return node id of named node (required to set node attributes) 
    @arg hdist ptr to distribution model
    @arg name  node identifier string 
    @return node ID */
int hdist_nodeid(const HDIST * const hdist, const char *name){
  ATT* att;
  assert(hdist && name);
  att=hdist->att;
  return(att_valid(att, name));   /* -1 is code for node not found */
}

/*--------------------------------------------------------------------------*/

/** find the string label associated with given node ID */
const char* hdist_nodename(HDIST * const hdist, int nodeid){
   assert(hdist);
   return att_valname(hdist->att,nodeid);
}

/*--------------------------------------------------------------------------*/

/** will return ptr to original string - use for reading only */
const char* hdist_getnodelabel(HDIST *hdist, int nodeid){
   ATT* att;
   HD_NODE* node;
   assert(hdist);
   att=hdist->att;
   if ((nodeid<0) || (nodeid>att->valcnt)) return NULL;
   node=hdist->node[nodeid];
   return node->label;
}

/*--------------------------------------------------------------------------*/

/** delete, create or change node labels */
int hdist_setnodelabel(HDIST *hdist, int nodeid, const char * label){
   ATT* att;
   HD_NODE* node;
   int llen;
   char* newlabel;
   assert(hdist);
   att=hdist->att;
   if ((nodeid<0) || (nodeid>att->valcnt)) return 0; /* failure: illegal id */
   node=hdist->node[nodeid];
   if (label){ 
      llen=strlen(label)+1;
      newlabel=malloc(llen*sizeof(char));  /* reserve memory for new label */
      if (!newlabel) return 0;             /* fail: could not reserve mem  */
      strncpy(newlabel,label,llen);        /* otherwise write a copy to it */
   } else {
       newlabel=NULL;                     /* label arg is NULL - just delete */
   }
   if (node->label) free(node->label);   /* clear any pre-exisiting lables */
   node->label=newlabel;                  /* .. and set pointer to new one  */
   return 0;
}
   
/*-----------------------------------------------------------------------------
  functions for loading and saving models
-----------------------------------------------------------------------------*/

/*  auxiliary function to hdist_load read node description data */
int _hdist_load_node(HDIST* hd, SCAN* scan, HD_NODE* node){
  int nid;
  ATT *att;   /* node name attribute */
  assert(hd && scan && node);
  att=hd->att;
  nid=node->valid;
  if (sc_token(scan)=='['){
    GET_CHR('[');
    if (sc_token(scan) != T_ID) return hdist_warn(HDW_LOADNODE);
    hdist_setnodelabel(hd,nid,(sc_value(scan)));    /* set label if present */
    /* string assumed to be in quoted for as it is  always written that way  */
    GET_TOK(); GET_CHR(']');
  }
  /* read the numerical values */
  if (sc_token(scan)!=T_NUM) return  hdist_warn(HDW_LOADNODE);
  node->flags=(int)(strtod(sc_value(scan),NULL)); GET_TOK();
  if (sc_token(scan)!=T_NUM) return  hdist_warn(HDW_LOADNODE);
  node->value=strtod(sc_value(scan),NULL); GET_TOK();
  if (sc_token(scan)!=T_NUM) return  hdist_warn(HDW_LOADNODE);
  node->covsv=strtod(sc_value(scan),NULL); GET_TOK();
  if (sc_token(scan)!=T_NUM) return  hdist_warn(HDW_LOADNODE);
  node->svrate=strtod(sc_value(scan),NULL); GET_TOK();
  /* no additional tests for content validity */
  return 0;
}

/*---------------------------------------------------------------------------*/

/* load annotated tree representation from save file, used internally by 
   hdist_load */
int _hdist_load_tree(HDIST* hd,SCAN* scan){
   int i,nid;
   HD_NODE *node;       /* current node */
   HD_NODE *p;          /* temporary node pointer for linking */
   assert( hd && scan );
   node=hd->root; p=NULL;
   if (sc_token(scan)!=T_ID) return  hdist_warn(HDW_LOADNODE);
   GET_TOK();          /* consume the root token */  
   do {
      if ((i=_hdist_load_node(hd,scan,node))) return i; /* get node data, */
      assert(p!=node);
      if (p!=NULL) _linkedge(p,node);    /* the link node to parent */
                                         /* check if children to parse ... */
      if (sc_token(scan)=='{'){          /* if so: open a new level for node */
        GET_CHR('{'); p=node;            /* must be followed by child node */
        if (sc_token(scan)!=T_ID) return  hdist_warn(HDW_LOADNODE);
      } else { 
          while (sc_token(scan)=='}'){
            GET_CHR('}');                    /* consume parenthesis */
            if (p!=NULL) p=(p->parent);      /* go up one level     */
          } /* while */
      } /* if */
      if (p==NULL) break;                   /* no siblings for root node   */
      /* register named child or sibling node to be parsed in next run ... */
      if (sc_token(scan)!=T_ID) return  hdist_warn(HDW_LOADNODE);
      if ((i=hdist_addnode(hd, sc_value(scan)))) return i;/* (create node) */
      nid=att_valid(hd->att,sc_value(scan)); node=hd->node[nid];  
      GET_TOK();                          /* ... and consume the name token */
   } while ((p!=NULL) && (sc_token(scan)!=T_EOF));  /* until back at root   */
   GET_CHR(';');                         /* now read the final terminator   */
   return 0;


}

/*--------------------------------------------------------------------------*/

/* function internally used by hdist_load() to read model parameters from
   file */ 
int _hdist_load_params(HDIST* hd, SCAN* scan){
  const char* bmodenames[4]={"UNDEFINED","ALLCLD","NON_EXHAUSTIVE",\
                             "EXHAUSTIVE"};
  const char* param;                             /* current parameter name  */
  /* make buffer large enough to store any token the scanner may return     */
  char  value[SC_BUFSIZE];                       /* e.g. the version string */
  int  i;
  assert(hd && scan);                            /* check arguments */
  GET_TOK();                                     /* read initial token   */
  /* read hdist parameters                                               */ 
  while (sc_token(scan) != T_EOF){               /* read parameter name  */
    param=sc_value(scan);
    if  (strcmp(sc_value(scan),"hd_version") ==0){
       GET_TOK(); GET_CHR('=');         /* consume param name and eq. sgn */
       strncpy(value,sc_value(scan),sc_len(scan)+1); /* save param value  */
       GET_TOK();                                    /* ... consume value */   
       GET_CHR(';');                                 /* ... and the term  */
    }
    else if (strcmp(sc_value(scan),"branch_mode") ==0){
      GET_TOK(); GET_CHR('=');
      hd->bmode=0;                    /* set initial branch mode to unknown */
      for (i=1;i<4;i++){              /* test known branch mode names       */
        if (strcmp(sc_value(scan),bmodenames[i])==0) hd->bmode=i;
      }
      if (hd->bmode<1 || hd->bmode>3){/* unrecognized or invalid branch mode */
        hdist_warn(HDW_LOADMOD,sc_value(scan));
      }
      GET_TOK(); GET_CHR(';');                   /* consume mode name and ; */ 
    }
    else if (strcmp(sc_value(scan),"lcorr") ==0){
      GET_TOK(); GET_CHR('='); 
      hd->lcorr=strtod(sc_value(scan),NULL);
      if (errno||(hd->lcorr<0.0)) hdist_warn(HDW_LOAD_LCO);
      GET_TOK(); GET_CHR(';');
    }
    else if (strcmp(sc_value(scan),"model") ==0){
      GET_TOK();GET_CHR('=');
      att_rename(hd->att,sc_value(scan));        /* read attribute */
      GET_TOK();
      i=_hdist_load_tree(hd,scan); 
      if (i>0) return(i);                        /* if something went wrong */
    } else {                                     /* unrecognized param. */
      hdist_warn(HDW_LOADPAR,sc_value(scan));
      GET_TOK(); GET_CHR('=');                   /* skip unrecognized  */ 
      GET_TOK(); GET_CHR(';');                   /* params             */ 
    }
  }
  return 0;                                      /* no parse error */
}

/*--------------------------------------------------------------------------*/

/** load a previously saved  hierarchical distribution from file
    @arg scan file scanner configured for reading 
    @return hierarchical distribution model
*/
HDIST * hdist_load(SCAN *scan){
  HDIST *hd;                                     /* the model to be created  */
  assert(scan);                                  /* check scanner argument   */
  hd=hdist_create("unnamed");                    /* will be renamed later on */
  if (!hd) {                                     /* check success            */
    hdist_warn(HDW_NOMEM);                       
    return (NULL);
  }
  if (_hdist_load_params(hd,scan)!=0) {          /* run parsing */
    hdist_delete(hd);                            /* on error cleanup and ...*/
    return NULL;                                 /* signal failure */
  }
  return(hd);                                    /* incomplete loading */
}

/*--------------------------------------------------------------------------*/

/** save distribution model and and associated information to file 
    @arg hdist prt to hierarchical distribution model
    @arg saveto file pointer 
    @return 0 on success */
int hdist_save(const HDIST *hdist, FILE *saveto){
   ATT*        att;
   const char* value;
   const char* bmodenames[]={"UNDEFINED","ALLCLD","NON_EXHAUSTIVE",\
                             "EXHAUSTIVE"};
   const char* ll;                     /* of opening and closing bracket... */
   const char* lr;                     /* enclosing optional label */
   const char* nname;                  /* name of current node */
   const char* lab;                    /* label of current node */        
   HD_NODE* p;                         /* for traversing nodes */
   int indent=0;                       /* indentation for tree writing */
   int i; 

   assert(hdist);                      /* check arguments ok */
   assert((hdist->bmode)>0 && (hdist->bmode)<=HDBM_SPLIT);
   att=hdist->att;
   fprintf(saveto,"/* save file for hierarchical distribution */\n"); 
   value=bmodenames[hdist->bmode];
   fprintf(saveto,"hd_version=%s;\n",HD_VERSION);   /* library version used to 
                                                          create savefile   */
   fprintf(saveto,"branch_mode=%s;\n",value);       /* branch mode */
   fprintf(saveto,"lcorr=%.12f;\n",hdist->lcorr);   /* laplace correction   */
   fprintf(saveto,"model=\"%s\"\n",att->name);      /* no terminating ';'   */
   /* write the full tree (including hidden values)                         */
   p=hdist->root;
   do {
     if (p->label!=NULL) {                  /* if optional label present    */
       lab=p->label; ll=" [\""; lr="\"]";   /* prepare extra output ...     */ 
     } else {
       lab="";ll="";lr="";                  /* ...prevent output otherwise  */
     }
     for (i=0;i<indent;i++) fprintf(saveto,"\t"); /* fancy indentation */
     if (p->flags&HDNF_ROOT) nname="<root>";
     else if (p->flags&HDNF_SV) nname="<set-valued>";
     else nname=att_valname(att,p->valid);
     assert(nname); /* now write the nodes : first name and flags, then ... */
     fprintf(saveto,"\"%s\"%s%s%s %i",nname,ll,lab,lr,p->flags); /*...rates */
     fprintf(saveto," %.12f %.12f %.12f",p->value,p->covsv,p->svrate); 
     /* get the next node and adapt ouput*/;
     if (p->child!=NULL) {         /* are there any children notes to write  */
       fprintf(saveto," {\n"); indent++;    /* start new level, indent ...   */
       p=p->child;  continue;               /* ...and continue with children */
     } /* no child found: seek siblings, then siblings of ancestors          */
     for (;p!=NULL && !(p->flags&HDNF_ROOT);){
       if (p->link!=NULL) {                   /* proceed with sibling, else  */
         p=p->link; break;
       } 
       fprintf(saveto,"\n");
       for (i=--indent;i>0;i--) fprintf(saveto,"\t");/* ... reduce indent...*/
       p=p->parent;                /* ...then cont with siblings of parents */
       fprintf(saveto,(p==NULL||p->flags&HDNF_ROOT)?"};":"}");     /* close */
     }  
     fprintf(saveto,"\n");
   } while(p!=NULL&& !(p->flags&HDNF_ROOT));     /* stop when back at root */
   return 0;                                     /* indicate successful save */
}
/*---------------------------------------------------------------------------*/

/** clear frame specifications from hdist 
@arg hdist ptr to hierachical distribution */
void hdist_clearframe(HDIST *hdist){
  HD_NODE* p;
  assert(hdist);
  for (p=hdist->root;p!=NULL;p=_nextdf(p,1))  /* just traverse all nodes   */ 
    p->flags&=~HDNF_NOEXPAND;                 /* ... and clear frame marks */
}
   
/*---------------------------------------------------------------------------*/

/** read a frame specification from file and set it as interpretation frame 
 @arg hdist ptr to hierarchical distributrion
 @arg nid numerical node identifier (obtained e.g. via from hdist_nodeid)
 @return status  code (0 for no error)
 @note This merely sets the node expansion flag - the node will still be 
       screened if the expansion path from the root node is interrupted.  
       Thus adding any ancestor will still prevent the node from actually 
       appearing in the active frame. */
int hdist_add2frame(HDIST* hdist, int nid){
  ATT* att;
  HD_NODE* node;
  assert (hdist);
  att=hdist->att;
  if ((nid<0) || (nid>att->valcnt)) return -1; /* failure: illegal id */
    node=hdist->node[nid]; 
  node->flags|=HDNF_NOEXPAND;
  return 0;
}

/*---------------------------------------------------------------------------*/  

/** show finest nodes in current frame on stderr */
void hdist_showframe(HDIST* hd){
   ATT* att;
   HD_NODE* p;
   assert(hd),
   att=hd->att;
   for (p=hd->root;p!=NULL;p=_nextdf(p,!((p->flags)&HDNF_NOEXPAND))){
     /* if (p->flags|HDNF_ROOT) continue;*/
     if ((p->flags)&(HDNF_NOEXPAND|HDNF_LEAF)){
        fprintf(stderr,"%s ",hdist_nodename(hd,p->valid));
     }
   }
   fprintf(stderr,"\n");
} 

/*-----------------------------------------------------------------------*/
/*   debug-only test functions                                           */
/*-----------------------------------------------------------------------*/
