/*-----------------------------------------------------------------------------
File    : hdist.h
Contents: data structure for dealing with set-valued data with hierarchical
          attribute domain
Author  : Frank Rgheimer
History : 20.02.2007 file ptree.h created
        : 23.02.2007 probability tree components 
        : 25.02.2007 added hdist interface 
        : 25.02.2007 switched to using attset.h for attribute management
        : 26.02.2007 added reference frame and valnametoids
        : 26.02.2007 fixed makefile
        : 27.02.2007 added probabilistic conversion between refernce frames
        : 19.08.2007 read tree structure from external files
	: 21.08.2007 added reading forest structure from external files
        : 22.08.2007 added dot-like graph language support :
        : 28.08.2007 switched to node based tree representation 
        : 31.08.2007 adapted parsing and tree construction to new nodes
        : 04.09.2007 support for set-valued distributions added
        : 08.09.2007 added branch modes for treating non-leaf nodes
        : 09.09.2007 hdist.h created for hierarchical domain management
                     modev hdist-structure from ptree.h to hdist.h
        : 10.01.2009 added lcorr field for laplace correction
        : 12.01.2009 introduced hdist_nodefreq() and hdist_nodecov() functions
        : 13.01.2009 added svrate flag for simplified sv-model
          18.01.2009 added hdist_setlogl() function
          08.02.2012 minor improvements to source documentation
          03.03.2012 minor improvements to source documentation replaced debug
                     macros improved variants form frdebug.h 
          15.03.2012 added HDQF flag group
          2012-05-10 changed application of Laplace correction to svrate to 
                     prevent unassiged probability mass in refinements with
                     only one child node and large Laplace corrections
                     -> new minor version as outside behavior changes
          2012-17-10 done testing release version interface 
-------------------------------------------------------------------------------
 Copyright 2007, 2009, 2012 Frank Rgheimer
 This program is distributed under the terms of the GNU Lesser General Public
 License.
-----------------------------------------------------------------------------*/

#define SC_SCAN     /* required for enabling  parser definitions */

#include <assert.h>
#include <errno.h>
#include <stdio.h> 
#include <stdlib.h>
#include <string.h>
#include <assert.h>
#include <stdarg.h>
#include <math.h>
#include "tabscan.h" /* Christian Borgelt's lib for parsing tables */
#include "attmap.h"  /* .. attribute management                    */
#include "parse.h"   /* .. and parsing/scanning text files         */ 
#include "frdebug.h"

/*---branch mode flags (aliases are grouped)---*/
#define HDBM_NONE 0       /* none: no mode selected */

#define HDBM_CONJ 1       /* conj: conjunctive interpretation (bad term)  */
#define HDBM_IMPR 1       /*       assigning a non-leaf node */

#define HDBM_DISJ 2       /* disj: disjunctive interpretation (bad term) */
#define HDBM_NXHTV 2      /* non-exhaustive definition of branching, i.e. 
concept space covered by child nodes does not exhaust concept space covered by 
parent nodes. This allows to assign probability mass to the parent node itself,
rather than any of its any singleton or multi-valued refinements, effectively
implementing an implicit "other" category at each branching point. Assigning
a observation to the more general term can be used to indicate that none of
the child terms applies and that the parent term is the most specific term that 
fully contains the desired concept (default for GO) */
#define HD_VERSION "2.3.1"  /* current version of the hdist library */

#define HDBM_SPLIT 3      /* split: split like dist. of  other cases, i.e, all 
unassigned prob. mass is prop. distributed to leaf nodes within local subtree */
#define HDBM_UCRT 3       /*        the union of the child concepts */
#define HDBM_XHTV 3       
/* completely exhaustes the concept space covered by their parent node.
 annotating a case with any non-leaf node indicates uncertainty about the true
 category on the higher level of detail. Untouched by this the child concepts 
 may overlap                                                                 */ 
 
/*---hdist mode constants---*/
/* not used - as probabilistic case is included in others */
#define HDM_NORMAL 0     /* standard probablilistic mode */
#define HDM_SETV 1       /* enable fuctionality for set-valued inputs */


/*---hdist node flags---*/
#define HDNF_NORMAL 0
#define HDNF_ROOT   1        /* node is the root node */
#define HDNF_SV     2        /* node represents a set-valued outcome */
#define HDNF_LEAF   4        /* node is a leaf node (not currently used) */
#define HDNF_MARK   8        /* flag for temporarily marking nodes */
#define HDNF_NOEXPAND 16     /* node acts as unexpandible leaf node for queries
                                (e.g. elements of a query frame selection) */

/* -- hdist query flags (used by _loglsubtree) --*/
#define HDQF_FROMSET 1 /* indcate queried subtree was reached via set inst. */
#define HDQF_STCUMUL 2 /* query likelihood of instantiation and all its re- 
                          finements rather than specific instantiation      */
#define HDQF_FRAME   4 /* evaluate relative to reference frame encoded in  
                          HDNF_EXPAND status */ 
extern double EPSILON;    /* epsilon  value for small positives float */
extern double TOLERANCE;  /* used in float comparisons */
/*---structs---*/

/** node of hierarchical distribution tree */
typedef struct _hd_node { /* node in hierarchical distribution */
  int flags;
  int valid;                /**< ID of the node == attribute value id */
  int nchildren;            /**< number of child nodes not counting sv-mode */
  struct _hd_node *parent;  /**< link to parent node */
  struct _hd_node *child;   /**< link to first child node (svgroup excluded) */
  struct _hd_node *svgroup; /**< special link to node for ext. set-valued mode 
                                 currently unused in all branch mode implem. */
  double value;          /**< cond. node probability|parent */
  double covsv;          /**< for coverage factor if set valued*/
  double svrate;         /**< rate of set-valued events in simple sv-mode*/ 
  struct _hd_node *link; /**< to link to next sibling */
  char*  label;          /**< node label string */
} HD_NODE;
    
/** hierarchical distribution model over structured attribute domain */
typedef struct _hdist{    
  ATT *att;        /**< attribute in questions */
  int vsz;         /**< size of internal vectors */
  int bmode;       /**< configure branch mode */
  int opstatus;    /**< set to indicate certain warning states, e.g. */
  HD_NODE *root;   /**< pointer to the root node*/
  HD_NODE **node;  /**< array of direct node pointers for idx-based access */
  double lcorr;                   
} HDIST;


/*---function declarations---*/

/* hdnode access functions */
int hdist_nodeid(const HDIST *hdist, const char *name);
const char* hdist_nodename(HDIST * const hdist, int nodeid);
const char* hdist_nodelabel(HDIST * const hdist, int nodeid);
int hdist_setnodelabel(HDIST *hdist, int nodeid,const char *  label);


/* hdist functions */
HDIST* hdist_create (const char *name); 
HDIST* hdist_parse(SCAN *scan, int bmode);
void hdist_delete (HDIST *hdist);
int hdist_resize (HDIST *hdist,int size);
/* int hdist_valadd (HDIST *hdist, const char *name, const char *parname); */
int hdist_addnode (HDIST *hdist, const char *name); 
int hdist_addedge (HDIST *hdist, const char *pname, const char *cname);
int hdist_valcnt(const HDIST *hdist);
void hdist_print(HDIST* hd);
int hdist_getfreqs(HDIST *hdist, TABSCAN *scanner, FILE* datafile);
/* query distribution and compute log-likelihoods  of samples */
double hdist_nodefreq(HDIST *hdist, HD_NODE *p);
double hdist_nodecov(HDIST *hdist, HD_NODE *p); 
/** removed const specifier from setlogl declaration -- compiler would not
    take char** argument even for const char * const * (should be safe) */
double hdist_setlogl(HDIST *hd,const char **elements, int maxel, 
                     const char ** failpos, const int flags);
/* functions for loading and saving models from/to file  */
/* note: currently configured frames are not saved! */
HDIST* hdist_load(SCAN * scan);
int hdist_save(const HDIST *hdist, FILE* saveto);
/* do not expand node for evaluation (make it an element of a frame) */
int hdist_add2frame(HDIST* hdist, int nid);
/* reset frame information to full expansion */
void hdist_clearframe(HDIST *hdist);
/* print  nodes of active frame to stderr */
void hdist_showframe(HDIST *hdist);

