/*-----------------------------------------------------------------------------
File    : psvmodel.c
Contents: various methods to generate and test models for set-valued 
          distributions; used to assess and compare representations for the 
          evaluation chapter of my dissertation
Author  : Frank Rgheimer
History : 15.12.2008 file created (based on ptree)
          16.12.2008 adapted command line processing from ptree
          21.12.2008 implemented file parsing for model construction
                     bug in parsing causes segfault with some artf. data
          29.12.2008 compiled on different machine - bug disappared (compiler
                     related?)
          29.12.2008 added readnext_sv() to separate parsing from model constr.
          30.12.2008 modified readnext_sv() to facilitate counting records
          02.01.2009 implement laplace correction
          03.01.2009 added -l option for setting laplace correction
          03.01.2009 implemented calculation of smple log-likelihood
          04.01.2009 switched to reading domain definition from .dom file
          05.01.2009 laplace correction now deactivated by default
          06.01.2009 added caluculation of upper and lower prob errors for model
                     vs. sample data
          08.01.2009 extended command line parameters to handle iprob and dist
                     modes, deactivated -i option to skip table headers
          18.01.2009 added testhd() for empirical eval. of hdist models
          06.05.2009 modified output format for easier parsing
                     added mean sample log likelihood (log of geometrical mean)
          08.05.2009 Initialized scanner states in readnext_sv() and 
                     imodel_induce() to most general one to tolerate comment
                     lines in data files (not extensively tested yet)
                     removed verbose outputs from productive version
          2010-10-01 adapted to new table package
          2012-02-13 fixed debug mode warning due to incorrect assert
          2012-02-05 changed response to zero probability events from error
                     to hint . 
          2012-03-20 adapted call to new signature of hdist_setlogl 
          2012-05-17 added explicits casts from (constant) char** to more
                     stringent const char** to suppress warnings from overly 
                     sensitive compiler 
-----------------------------------------------------------------------------*/

/*---includes---*/
#include <errno.h>
#include "psvmodel.h"
#include "hdist.h"

/*---definitions---*/

#define DESCRIPTION "various methods of modelling set-valued distributions\n"
#define VERSION "\
version 0.7.0  (2012-03-20)            \
\n         (c) 2007-2011,2012 Frank Rgheimer\n\
attmap, tabscan and parser modules     \
\n         (c) 1996-2010 Christian Borgelt\n"

/*---error codes for main program---*/
#define E_NONE  0         /*no error*/
#define E_OPENF 1         /*failed to open file */
#define E_EOF 2           /*unexpected EOF*/
#define E_ATTVAL 3        /*unknown attribute value*/
#define E_READFILE 4
#define E_OPTION 5        /* unknown option*/
#define E_ARGCNT 6        /* invalid number of arguments */
#define E_NO_MEM 7        /* out of memeory */
#define E_PARSETREE 8     /* failure to parse tree: */
#define E_PARSETBL 9      /* error parsing table file: */
#define E_OPTARG 10       /* missing option argument */
#define E_BMODE 11        /* invalid branching mode option argument: %s */
#define E_OPTARG_M 12     /* invalid argument for -m option: %s */
#define E_LOGZERO 13      /* zero prob. while computing log-likelihood*/
#define E_LCORR 14        /* invalid value for laplace correction */
#define E_PARSE 15        /* error while parsing file: %s */
#define E_UNKNOWN 16      /* unknown error */

/*---error messages---*/


/* error must be called with matching parameters. Failure to to so will
   usually trigger a segmentation fault                                  */
static char *errormsg[]={
  /*E_NONE 0*/    "no error\n",
  /*E_OPENF 1*/   "failed to open file %s\n",
  /*E_EOF 2*/     "unexpected end of file %s\n",
  /*E_ATTVAL 3*/  "encountered unexpected attribute value while parsing\n",
  /*E_READFILE 4*/"error reading file\n",
  /*E_OPTION 5*/  "unknown option %s \n",
  /*E_ARGCNT 6*/  "invalid number of arguments\n",
  /*E_NOMEM 7*/   "out of memory\n",
  /*E_PARSETREE 8*/ "failure to parse tree: %s\n",
  /*E_PARSETBL 9*/"error parsing table file: %s\n",
  /*E_OPTARG 10*/ "missing option argument\n",
  /*E_BMODE 11*/  "invalid branching option argument: %s\n",
  /*E_OPTARG_M 12*/ "invalid argument for -m option: %s\n",
  /*E_LOGZERO 13*/ "cannot compute log of likelihood 0 - try \
enabling laplace correction (-l option)\n",
  /*E_LCORR 14*/  "invalid value for laplace correction\n",
  /*E_PARSE 15*/ "error while parsing file: %s\n",
  /*E_UNKNOWN */ "unknown error\n"
};

/* improving readibility */
#define FALSE 0
#define TRUE 1

/*---model type codes---*/
#define MTC_IPROB 0         /* independent boolean variables for set mbrshp  */
#define MTC_HDIST 1         /* use hierarchical distribution model           */
#define MTC_UNKNOWN 2       /* unknown mode, e.g. parameter from newer vers. */

/*---globals---*/
static char   *progname;
int	      options=0;     /* for option flags, e.g. set-valued eval */
static TABSCAN *scanner= NULL; /* scanner for table file */
static SCAN   *scan   = NULL; /* general scanner */  
static HDIST  *hd     = NULL; /* for representation of distributions      */
static ATTSET *attset = NULL; /* for .dom-file parsing                    */
static FILE *datafile=NULL; /* a pointer to a data file which constist of a 1 
                     line header with <space>-separated attribute names and
                     one line for each  record with <space>-separated attribute
                     values. Used as training data for model induction    */
static FILE *testdatafile=NULL; /* ptr to test data file for evaluation
                                   exprmnt. see above for format          */   
char withheader= FALSE;      /* assume file has no header by default      */
double lcorr=0.0;            /* laplace correction default value          */
/* static FILE *treefile;    used to read tree structure - format is subset of
                     graphwiz dot language see example.htr  
                     only required for models with structured domain 
                     (currently not used) */

/*---------------------------------------------------------------------------*/
/*  general purpose functions for argument & error handling                  */
/*---------------------------------------------------------------------------*/

/*  Option arguments (model type selection)                                  */

typedef struct {      /* for comfortable access to option params */
  int code;           /* numeric code */
  char* name;         /* name */
} MINFO;              /* from C.B. */


static const MINFO modeltype[] = {
  {MTC_IPROB, "iprob"},   /* probab. model with independent set membership */
  {MTC_HDIST, "hdist"},   /* hierarchical sv-distribution model */
  { -1, NULL}             /* invalid code (sentinel) */
};
/* clean up  all remaining global objects and mark them as deleted to    */
/* simplify cleanup all pointers to invalid objects should be set to NULL*/
static void cleanup(){
  if (scanner) ts_delete(scanner); scanner=NULL;
  if (scan) sc_delete(scan); scan=NULL;
  if (hd) hdist_delete(hd); hd=NULL;
  if (attset) as_delete(attset); attset=NULL;
  /* if (treefile) fclose(treefile); treefile=NULL; */
  if (datafile) fclose(datafile); datafile=NULL;
  if (testdatafile) fclose(testdatafile); testdatafile=NULL;
}


static int code (const MINFO *tab, const char *name)
{                                  /* map string to integer code */
  for (;tab->name!=NULL; tab++)    /* try names from table */
    if (strcmp(tab->name,name)==0)  
      return tab->code;            /* return code if found */
  return -1;                       /* nothing was found */
}



/* some error messages have arguments; see header file for individual 
   message format; failure to supply these will usually trigger a
   segmentation fault                                                    */

static void error (int code, ...){
  va_list args;                    /* the variable argument stuff */
  const char* msg;                 /* the error message */
  if (code>E_UNKNOWN) code=E_UNKNOWN;
  if (code>0){
    msg=errormsg[code];
    if (!msg) msg = errormsg[E_UNKNOWN];
    fprintf(stderr,"\n%s: ",progname);
    va_start(args, code);          /* last argument before elipsis */
    vfprintf(stderr,msg,args);     /* print combined error message */
    va_end(args);                  /* finish processing varguments */
  }
  cleanup();                       /*clean up memory and exit program*/
  exit(code);
}

/*-----------------------------------------------------------------------------
   Model data structures and associated functions                            
-----------------------------------------------------------------------------*/

/* model distribution by treating the occurence of individual attribute values
   in as independent boolean variables */  

void imodel_delete(IMODEL* imodel){
  assert(imodel);
  if (imodel->att) att_delete(imodel->att);
  if (imodel->value) free(imodel->value);
  free(imodel);
}

/* induce model from data - return NULL if anything goes wrong 
 based on hdist_getfreqs, but more flexible. Adds new nodes when encountered */

IMODEL* imodel_induce(TABSCAN* scanner, FILE* trndata, ATT* eatt){
   /* variables for target data structure */
   IMODEL* imodel;             /* for building the structure to be returned */
   ATT* att;                   /* pointer to internal attribute copy        */
   char ignrecord=FALSE;       /* if the file contains a header, the first  */
                               /* non-comment record must be ignored        */
                               /* out of use - maybe I will re-use it later */
   /* variables used on record level and below */ 
   char buffer[2048]; /* create a large reading buffer on the stack for long */
   const int bufsize=2048;
   int d=TS_FLD;  /* used for scanner state; start with most general one     */
   /* variables used in parsing set descriptions */
   int l;
   int elcnt;   /* elements already read if parsing a set                    */
   int status;  /* return status for attribute function calls */ 
   char *token; /* for extracting tokens from a set                          */
   char *token2;/* set to follow_up token when parsing strings               */
   double cnt;
 
   /* initialize internal data structures */ 
   assert(scanner&&trndata&&eatt);                          /* check inputs */
   imodel=(IMODEL*)(malloc(sizeof(IMODEL)));   /* reserve memory for result */
   if (!imodel) return NULL;                                  /*reserve mem */
   imodel->reccnt=0;                              /* create data structures */
   imodel->att=att=att_clone(eatt);                       /* copy attribute */
   if (!att) {imodel_delete(imodel); error(E_NOMEM);} 
   imodel->vsz=att->valvsz;
   imodel->value=(double*)malloc(imodel->vsz*sizeof(double));
   if (!imodel->value) {
     imodel_delete(imodel);  
     return NULL;
   }
  
  /* processing the records of the training data file */     
  do {                                   /* read records from file          */
    /* if(tfs_skip(scanner,trndata)!=0) continue; *//* ignore comment lines */   
    if (ignrecord) {                     /* if the file has a header...     */
       while (ts_next(scanner,trndata,NULL,0)==TS_FLD);     /* do not eval  */
       ignrecord=FALSE;                  /* ... the 1st non-comment record. */
    }
    do {                                 /* ignore all fields but the last  */
      d = ts_next(scanner,trndata,buffer,bufsize);    /* one in the record  */
    } while (d == TS_FLD);
    if (d<0) {                           /* something is wrong with the file*/
       imodel_delete(imodel);            /* clean up the heap structures    */
       return(NULL);                     /* return and indicate failure     */
    } 
    /* loop to process the set of values in the current field */
    elcnt=0; token=token2=buffer;
    if (*token=='{') token2=(++token);   /* skip leading '{' if present */
    for (;*token!='\0';token=token2) {   /* while there is something to read*/
      while((*token2!='\0')&&(*token2!='}')&&(!ts_istype(scanner,TS_BLANK,\
*token2)))
        token2++;                        /* seek end of first token */
      if (*token2!='\0') {
	*token2='\0';                    /* terminate token and skip blanks */
        for (token2++;ts_istype(scanner,TS_BLANK,*token2);token2++);
      }
      l=att_valid(att,token);            /* determine its index code value */
      if (l<0) {                         /* encountered unregistered value */ 
        /* register node and update data structures */
        /* resize data structures if neccessary */
        fprintf(stderr,"warning: adding unregistered attribute value %s\n"\
,token);
        status=att_valadd(att,token,NULL);
        if ((imodel->vsz)<(att->valvsz))
          imodel->vsz=att->valvsz;
          imodel->value=(double*)realloc(imodel->value,\
imodel->vsz*sizeof(double));
        if ((status!=0) || (!imodel->value)) {
           imodel_delete(imodel); 
           return(NULL);
        }
        l=att->inst.i;                  /* read index of newly created value */
        imodel->value[l]=0;             /* ... and init the resp. counter */
      }
      /* increase the occurance counters */
      imodel->value[l]++;
      elcnt++;
    } /* nodes in the set are marked */
    if (elcnt>0) {                       /* ignore empty lines & empty sets */
      imodel->reccnt++;
    }
  } while (d > TS_EOF);                 /* stop when eof is reached        */
  /* estimate coverage probability using laplace correction */
  for (l=0;l<att->valcnt;l++){
    cnt=imodel->value[l];
    imodel->value[l]=(cnt+lcorr)/(imodel->reccnt+(2.0*lcorr));
  }
  return(imodel);
}

void imodel_print(IMODEL* imodel){
  int i;
  ATT* att;
  const char* label;
  double cov;
  assert(imodel);
  att=imodel->att;
  for (i=0;i<att->valcnt;i++){
    label=att_valname(att,i);
    cov=imodel->value[i];
    printf("%s  \t %f\n",label,cov);
  }
  return;
}

/* load test data and evaluate model*/
/* evaluation based on log-likelihood of sample - the higher the likelihood*/
/* of the sample, the better the generalization of the test data by the mdl*/
int imodel_test(IMODEL* imodel, TABSCAN* scanner, FILE* tstdata){
  ATT* att;          /* shortcut */
  int elcnt,reccnt,d,maxel,l,i;
  double logp_mt;    /* log likelihood of the empty set as sample realization*/
  double logp_smpl;  /* log likelihood of observed sample S */
  double logp_test;  /* log likelihood od the test data */
  double logp_msmpl; /* mean log likelihood of sample in test data */
  double loglikabs; 
  double qerrfrq,qerrcov;    /* divergence of estimated and observed rates
                                using squared error sum measures */          
  char buffer[2048];         /* reading buffer for parser */
  int  bufsz=2048;
  char *valnames[100];       /* for storing string pointers */
  char **p;                  /* for traversing token stings */ 
  char* token;
  double* frq_cov_o  =NULL;   /* observed element coverage rates in sample  */
  double* frq_sglt_o =NULL;   /* observed frequency of singleton outcomes */
  double* logpin     =NULL;   /* precomputed values 
                                 log(P(e \in S))-log(P(e \not\in S))  
                                 The sample log-likelihood is computed by adding
                                 the respective value for all elements in the 
                                 sample to the log-likelihood of the empty set 
                                 instantitation*/
  double* pcov;               /* array with coverage probabilities */

  maxel=(sizeof(valnames)/sizeof(char*))-1;   /* =99 */
  assert(imodel&&scanner&&tstdata); /* check inputs and define shortuts*/
  att=imodel->att; assert(att);     
  pcov=imodel->value; assert(pcov);
  /* precompute tables for log-likelihood to speed up calculation */
  logpin=(double*)malloc(sizeof(double)*att->valcnt); /* allocate table mem */
  frq_cov_o=(double*)calloc(att->valcnt,sizeof(double));
  frq_sglt_o=(double*)calloc(att->valcnt,sizeof(double));
  if (!(logpin&&frq_cov_o&&frq_sglt_o)) error(E_NOMEM);
  logp_mt=0;            /* likelihood of the sample of lenght 0 is 1 -> log=0 */     
  for (i=0;i<att->valcnt;i++){
    if (pcov[i]==0.0||pcov[i]==1.0) error(E_LOGZERO); 
    loglikabs=log(1-(pcov[i]));
    logpin[i]=log(pcov[i])-loglikabs;
    logp_mt+=loglikabs; /* accum. logL(e \not\in S) to compt. logL(S={}) */ 
  }
  /* printf("logp_mt: %f\n",logp_mt); */
  /* compute log-likelihood and upper, lower prob errors from test data */
  printf("-----------------\n");
  logp_test=0.0;
  reccnt=0;
  while ((d=readnext_sv(buffer,bufsz,valnames,maxel,scanner,tstdata))){
   if (d<0) break; /* exit loop with errorcode as return value */
   /* for each nonempty line compute the prob. of the observed set */
   logp_smpl=logp_mt; elcnt=0; reccnt++;
   for (p=valnames;*p;elcnt++) {
     token=*p; p++;
     /* printf("%s %s",token,*p?"& ":"\\\\\n"); */
     l=att_valid(att,token);
     if (l<0) {
       printf("warning: ignoring unknown attribute value: %s\n",token);
     }
     else {
       logp_smpl+=logpin[l];  /* modify sample probability of element */
       frq_cov_o[l]++;      /* update coverage and singleton counters */
       if (elcnt==0 && !(*p)) frq_sglt_o[l]++; /* cond. for singleton */
     }
   }
   logp_test+=logp_smpl; 
   /* printf("log likelihood of sample: %f\n",logp_smpl); */  /*for debugging */   
  }
  logp_msmpl=(reccnt>0)?(logp_test/reccnt):(1.0);

  /* compute realtive value freq and coverage rate in test data (no lcorr) */
  qerrfrq=qerrcov=(reccnt>0)?(0.0):(-1.0);
  for (i=0;(reccnt>0)&&(i<att->valcnt);i++){
     frq_sglt_o[i]=(frq_sglt_o[i]/reccnt);
     frq_cov_o[i]=(frq_cov_o[i]/reccnt);
     qerrfrq+=pow( exp(logp_mt+logpin[i])-frq_sglt_o[i]  ,2.0 );
     qerrcov+=pow( (pcov[i]-frq_cov_o[i])                ,2.0 );
  }
  printf("\n-------------------------------------------------------------\n");
  printf("model\t logL(D_test)\tlogL(d_avg)\tsum err_frq^2\
\tsum err_cov^2\n");
  printf("-------------------------------------------------------------\n");
  printf("iprob\t%f\t%f\t%f\t%f\n",logp_test,logp_msmpl,qerrfrq,qerrcov);
/*printf("iprob: \nlog-likelihood of test data &\t%f\n",logp_test);
  printf("err_{cov}^2 &\t%f\n",qerrcov);
  printf("err_{frq}^2 &\t%f\n",qerrfrq);*/
  #ifndef NDEBUG
 /* printf("-----------------\n");
  printf("observed on test data: \n");
  for (i=0;(reccnt>0)&&(i<att->valcnt);i++)
    printf("value, f_a(sglt), f_a(cov): %s  %f  %f\n",
att_valname(att,i),frq_sglt_o[i],frq_cov_o[i]);
  printf("-----------------\n");*/
  #endif
  if (logpin)     free(logpin);
  if (frq_cov_o)  free(frq_cov_o);
  if (frq_sglt_o) free(frq_sglt_o);
  return(d);
}

/*===========================================================================*/
/* evaluation of a hdist model */
int testhd(HDIST* hd, TABSCAN* scanner, FILE* tstdata){
  int      i,l,d,maxel,elcnt,reccnt;
  char     buffer[2048];       /* reading buffer for parser */
  int      bufsz=2048;
  char*    valnames[100];      /* for storing string pointers */
  char**   p;                  /* for traversing token stings */ 
  char*    token;
  char*    errtoken;           /* token for error message */
  char     zeroprob;           /* indicate db has proabilitiy 0 -- warn */
  double   logp_test;          /* to aggregate log-likelihood of test data */
  double   logp_smpl;          /* log likelihood of current sample */
  double   logp_msmpl;         /* mean log likelihood of test data sample */
  double   qerrfrq;            /* squared error sum (singleton frequencies) */
  double   qerrcov;            /* squared error sum (coverage rates) */
  double*  frq_cov_o  =NULL;   /* observed element coverage rates in sample  */
  double*  frq_sglt_o =NULL;   /* observed frequency of singleton outcomes   */
  ATT*     att;
  HD_NODE* pn;
  
  
  assert(hd&&scanner&&tstdata);
  att=hd->att; assert(att);  
  zeroprob=FALSE;   
  maxel=(sizeof(valnames)/sizeof(char*))-1; /* =99 */
  frq_cov_o=(double*)calloc(att->valcnt,sizeof(double));
  frq_sglt_o=(double*)calloc(att->valcnt,sizeof(double));
  if (!(frq_cov_o&&frq_sglt_o)) error(E_NOMEM);
  logp_test=0.0;               /* test with no sample always ok */
  reccnt=0;
  while ((d=readnext_sv(buffer,bufsz,valnames,maxel,scanner,tstdata))){
    if (d<0) break;            /* exit loop with errorcode as return value */
    errtoken=NULL; 
    /* update log likelhood and observed freq/coverage */
    logp_smpl=hdist_setlogl(hd, (const char **) valnames, maxel,
                                (const char **) &errtoken,0);
    /* using typecast to supress warnings of overly careful compiler */
    /* errtoken is reset too "" for mere zero probs */
    logp_test+=logp_smpl;
    if (errtoken){
        if (!*errtoken) { 
            zeroprob=TRUE; errtoken=NULL; /* sample has prob. 0, reset token */
        }
        else error(E_PARSE,tstdata);     /* failure reading test data */
    }
#ifndef NDEBUG
    printf("sample log-likelihood: %f\n", logp_smpl); 
#endif
    elcnt=0; reccnt++;
    for (p=valnames;*p;elcnt++) {
      token=*p; p++;
      /* printf("%s %s",token,*p?"& ":"\\\\\n"); */
      l=att_valid(att,token);
      if (l<0) {
        printf("warning: ignoring unknown attribute value: %s\n",token);
      }
      else {
        frq_cov_o[l]++;        /* update coverage and singleton counters */
        if (elcnt==0 && !(*p)) frq_sglt_o[l]++; /* cond. for singleton */
      }
    }
  }
  logp_msmpl=(reccnt>0)?(logp_test/reccnt):(1.0);

  /* compute realtive value freq and coverage rate in test data (no lcorr) */
  qerrfrq=qerrcov=0.0;
  for (i=0;(reccnt>0)&&(i<att->valcnt);i++){
     frq_sglt_o[i]=(frq_sglt_o[i]/reccnt);
     frq_cov_o[i]=(frq_cov_o[i]/reccnt);
     pn=hd->node[i]; assert(pn);
#if false
#ifndef NDEBUG
     printf("%s : (freq,obs)=(%f,%f), (cov,obs)=(%f,%f)\n",
        (att_valname(hd->att,i))?att_valname(hd->att,i):"n.a.",
        hdist_nodefreq(hd,pn),frq_sglt_o[i],
        hdist_nodecov(hd,pn),frq_cov_o[i]);
#endif
#endif
     qerrfrq+=pow( (hdist_nodefreq(hd,pn)-frq_sglt_o[i]) ,2.0 );
     qerrcov+=pow( (hdist_nodecov(hd,pn)-frq_cov_o[i]) ,2.0 );
  }
  printf("\n-------------------------------------------------------------\n");
  printf("model\t logL(D_test)\tlogL(d_avg)\tsum err_frq^2\
\tsum err_cov^2\n");
  printf("-------------------------------------------------------------\n");
  printf("hdist\t%f\t%f\t%f\t%f\n",logp_test,logp_msmpl,qerrfrq,qerrcov);
  /*
  printf("hdist: \nlog-likelihood of test data &\t%f\n",logp_test);
  printf("err_{cov}^2 &\t%f\n",qerrcov);
  printf("err_{frq}^2 &\t%f\n",qerrfrq);
  */
  if (logp_test==log(0.0)) fprintf(stderr,
     "Encountered samples with probability 0.0! Hint: rerun with option -l to"\
     "\nenable Laplace correction!\n");
  if (frq_cov_o)  free(frq_cov_o);
  if (frq_sglt_o) free(frq_sglt_o);
  return(errtoken?(-1):0); /* error status */
}

/*===========================================================================*/
  
/*-----------------------------------------------------------------------*/ 
/*--- definitions of common functions---*/
/*-----------------------------------------------------------------------*/



/* returns 1 if valid record is found, -1 on error, 0 if no record is left  */
int readnext_sv(char* buffer, int bufsize, char** tokens, int maxel,\
 TABSCAN* scanner, FILE* input){
   int d=TS_FLD;  /* for scanner state; indicate no previously read record  */
   /* variables used in parsing set descriptions */
   int elcnt,c;    /* elements already read if parsing a set; char from file */
   char *token;    /* for extracting tokens from a set                       */
   char *token2;   /* set to follow_up token when parsing strings            */
   assert(scanner&&input&&buffer&&tokens);                   /* check inputs */
   assert(maxel>=0);
   elcnt=0;
   tokens[0]=NULL;                        /* initialize as empty set */
   do {                                   /* read records from file          */
    c=getc(input);
    if (c == EOF) return 0; else ungetc(c,input);     /*test for end of file */
    /* if(tfs_skip(scanner,input)!=0) continue; *//* ignore comment lines    */
    /* found a non-comment record...                                         */
    do {                                 /* ignore all fields but the last.. */
      d = ts_next(scanner,input,buffer,bufsize);    /* ... one in the record */
    } while (d == TS_FLD);
    if (d<0) {                           /* something is wrong with the file*/
       return(d);                        /* return and indicate failure     */
    } 
    /* loop to process the set of values in the current field */
    elcnt=0; token=token2=buffer;
    if (*token=='{') token2=(++token);   /* skip leading '{' if present */
    for (;*token!='\0';token=token2) {   /* while there is something to read*/
      while((*token2!='\0')&&(*token2!='}')&&(!ts_istype(scanner,TS_BLANK,\
*token2)))
        token2++;                        /* seek end of first token */
      if (*token2!='\0') {
	*token2='\0';                    /* terminate token and skip blanks */
        for (token2++;ts_istype(scanner,TS_BLANK,*token2);token2++);
      }
      if (maxel<elcnt) return(ESV_ELCNT);/* check pointer array space */
      tokens[elcnt]=token; 		 /* register new token */
      elcnt++;
    }  /* nodes in the set are marked */
  } while ((d > TS_REC)||!(scanner->cnt));/* stop at end of non-empty record */
  tokens[elcnt]=NULL;                   /* invalidate following tokens ptrs  */
  return 1;
}

/*===========================================================================*/
    
 
/*---help screen---*/

static void help (void){
  fprintf(stderr,"\n");
  fprintf(stderr,"usage: %s [options] domfile trnfile tstfile\n", progname);
  fprintf(stderr,"%s\n", DESCRIPTION);
  fprintf(stderr,"%s\n", VERSION);       /* ...print usage message... */
  fprintf(stderr,"trnfile and tstfile contain tables with training and test\
data respectively.\n"); 
  fprintf(stderr,"Data files consist of optional comment part (lined starting\
with '#' followed\n");
  fprintf(stderr," by a header line with field names. Each data record corresp\
onds to one line,\n");
  fprintf(stderr,"with field being separated by tabulators.\n");
  fprintf(stderr,"Any field may either contain an identifier or a set of ident\
ifiers that are\n");
  fprintf(stderr,"separated by spaces and enclosed in curly brackets (e.g. {A \
B C}).\n");
  fprintf(stderr,"By default, the last field is assumed to contain the attribu\
te to be processed\n");
  fprintf(stderr,"list of supported options: \n");
  fprintf(stderr,"-m<MCODE> set model type used for approximation of set-  \n");
  fprintf(stderr,"          distribution to <MCODE>. Valid values are: \n"); 
  fprintf(stderr,"          iprob - model set membership as independent \n");
  fprintf(stderr,"                  binary variables - domfile as produded by \
dom\n");
  fprintf(stderr,"          hdist - use hierarchical distribution model \n");
  fprintf(stderr,"                  domfile is a directed tree in .dot format\
\n");
  fprintf(stderr,"          (default = iprop)\n");
  fprintf(stderr,"-i        use the first non-comment record of the input as \
header\n");
  fprintf(stderr,"          by default all records are assumed to be data\n");
  fprintf(stderr,"-l<float> set value for laplace correction (default=0.0)\n");
/* fprintf(stderr,"\n"); */
  return;
}

/*-----------------------------------------------------------------------*/


/*---main function---*/

int  main(int argc, char* argv[]){
  int i, k=0;                      /* loop variable and argument counter */
  char       *s;                   /* for scanning options */
  char       *mtoastr = NULL;      /* buffer for model type argument string */
  char       *lcoastr = NULL;      /* the laplace correction option argument*/ 
  char       **optarg = NULL;      /* for redirecting  option arguments     */
  char       *fn_trn  = NULL;      /* name of table file used for training  */
  char       *fn_tst  = NULL;      /* name of table file with test data     */
  char       *fn_dom  = NULL;      /* name of base attribute domain file    */
  int        mt_code  = MTC_IPROB; /* default mode for model type */
  IMODEL*    imdl     = NULL;      /* distribution model with indep. assumt.*/
  HDIST*     hd       = NULL;      /* hierarchical distribution model */
  ATT*       att      = NULL;      /* base attribute */


  progname=argv[0];                /* get program name for messages */
  /* --- print greeting message --- */
  if (argc > 1) {                  /* if there are arguments given*/   
    fprintf(stderr, "%s - %s\n", progname, DESCRIPTION);
    fprintf(stderr, VERSION); }    /* print greetings... */
  else{                            /* ...otherwise...*/ 
    help();                        /* ...show help message */
    exit(2);                       /* and abort program */
  }
   
  /* --- evaluate arguments --- */ 
  for (i=1; i< argc; i++) {        /* traverse the argument vector   */
    s=argv[i];                     /* get next argument */
    if (optarg) {                  /* if at start of an option argument  */
      *optarg=s;                   /* set resp. var to reading position  */
      optarg=NULL;                 /* doubles as an indicator tnx to C.B */
      continue;                    /* read next command line argument    */
    }
    /* otherwise */
    if ((*s == '-') && *++s) {     /* for options ignore leading '-' */
      while (*s) {                 /* may be list of options */
        switch (*s++) {            /* evaluate options */
        case '?': help();                  return 0; 
	case 'm': optarg=&mtoastr;         break;    /* prepare optional args */
        case 'l': optarg=&lcoastr;         break;    
        /* case 'i': withheader=FALSE;        break;    */  /* deactivated */
	default : error(E_OPTION,s-2);     return 0; /* unknown option error  */
	}
        /* option arguments may follow directly or in the next cmdarg */
        if (!*s) break;            /* end of cmdarg - try next one  */
        if (optarg) {*optarg =s; optarg=NULL; break;} 
        /* identify rest of current cmdarg with the option argument */
      }
    }
    else {                         /* process non-option arguments */
      switch (k++) {               /* match arguments by position */
        case  0: fn_dom=s; break;  /* use 1st argument for domain file name */
        case  1: fn_trn=s; break;  /* use 2st argument for trn-data file name */
        case  2: fn_tst=s; break;  /* and 3nd argument for tst-data file name */
      default: error(E_ARGCNT); break; /* too many arguments */
      }
    }
  }
  if (optarg) error(E_OPTARG);     /* missing option argument */
  /* match option argument strings to internal flags */
  if (mtoastr!=NULL) {
    mt_code=code(modeltype,mtoastr);  /* convert -m option arg. from string */
    if (mt_code<0) error(E_OPTARG_M,mtoastr);
  }
  if (lcoastr!=NULL) {              /* parse value for laplace correction */
    lcorr=strtod(lcoastr,NULL);
    if (errno||(lcorr<0.0)) error(E_LCORR);
  }
  if (k<3) error(E_ARGCNT);        /* check required arguments */
 
 
  /* build the models and run the tests ...*/ 
  assert(fn_dom);
  assert(fn_trn);
  assert(fn_tst);
  /* read domain file and create attribute */
  scan=sc_create(fn_dom);  if (!scan) {
    if (!fn_dom||!*fn_dom) error (E_NOMEM);
    else error(E_OPENF,fn_dom);
  }
  pa_init(scan);                /* use parser ext. for reading hdist */
  
  switch (mt_code){
    case MTC_IPROB:  
      /* parse the conventional domain file (6 lines copied from C.B.) */
      printf("\n...parsing domain file \t"); 
      attset=as_create("unnamed",att_delete); if (!attset) error(E_NOMEM);   
      if ((sc_nexter(scan)   <  0)  /* start scanning (get first token) */
      ||  (as_parse(attset, scan, AT_ALL) != 0)
      ||  (as_attcnt(attset) <= 0)  /* parse attribute set and */
      ||  !sc_eof(scan))            /* check for end of file */
      error(E_PARSE, sc_fname(scan));  
      att=as_att(attset,as_attcnt(attset)-1); /* use last attrib. as base att.*/
      printf(" - found %i values in base domain\n",att_valcnt(att));
      break;
    case MTC_HDIST:
      hd=hdist_parse(scan,HDBM_DISJ);     /* read the domain hierarchy -adapt */
      if (!hd) error(E_PARSETREE);
      break;
    default: error(E_UNKNOWN);
  }
  if (scan) sc_delete(scan); scan = NULL; /* delete the scanner */
  /* process input table */
  scanner=ts_create();
  if(!scanner) error(E_NOMEM);/* scanner initialization & config.*/
  ts_chars(scanner,TS_FLDSEP,",\t");/* set field separators, ...        */
  ts_chars(scanner,TS_RECSEP,"\n"); /* ...and record separator          */
  ts_chars(scanner,TS_COMMENT,"#"); /* set indicator for comment lines  */
  /* blank symbols serve to distinguish elements in sets */
  
  if (!(datafile=fopen(fn_trn,"r"))) error (E_OPENF,fn_trn);
  printf("\n...reading training data from file %s\n",fn_trn);
  switch (mt_code){
    case MTC_IPROB: 
      if (!(imdl=imodel_induce(scanner,datafile,att))) error(E_PARSETBL,fn_trn);
      fclose(datafile); datafile=NULL;  /* close the data file, mark closed */
#ifndef NDEBUG
      imodel_print(imdl);               /* show induced model */
#endif
      break;
    case MTC_HDIST:
      assert(hd);
      hd->lcorr=lcorr;
      if (hdist_getfreqs(hd,scanner,datafile)!=0) error(E_PARSETBL,fn_trn);
#ifndef NDEBUG
      hdist_print(hd);
#endif
      break;
    default: error(E_UNKNOWN); 
  }

  if (!(testdatafile=fopen(fn_tst,"r"))) error (E_OPENF,fn_tst);
  printf("\n...reading test data from file %s\n",fn_tst);
  ts_reset(scanner);
  switch (mt_code){
    case MTC_IPROB:  
      assert(imdl);  
      if (imodel_test(imdl,scanner,testdatafile)!=0) error(E_PARSETBL,fn_tst);
      break;
    case MTC_HDIST:
      assert(hd);
      if (testhd(hd,scanner,testdatafile)!=0) error(E_PARSETBL,fn_tst);
      break;
    default: error(E_UNKNOWN);
  }    
  fclose(testdatafile);  testdatafile=NULL; /* and close and prev. dbldel */
  cleanup();
  return 0;
}
